# Licensed under GPL version 3 - see LICENSE.rst
'''Tools for setting up instruments in the Rowland Torus geometry.

This includes an object that represents the Rowland torus itself (imaginatively
called `~marxs.design.rowland.RowlandTorus`), some helper functions used to set
the right parameters for the torus and a few classes that are derived from
`marxs.simulator.ParallelCalculated`, each placing elements such as gratings
or detectors on the Rowland torus. There are many ways to do that, e.g. the
limits can be defined in x,y,z coordinates or in theta, phi coordinates on the
torus, gratings can be ordered in concentric circles or pack densely in a
rectangular area etc. At this point, the different classes do not exploit all
these possibilities, they merely give a few of many possible ways to set up an
instrument.

These classes may be generalized in the future.
'''
import numpy as np
from scipy import optimize
import transforms3d

from ..optics.base import OpticalElement
from ..base import MarxsElement
from ..optics import FlatDetector
from ..math.utils import e2h, h2e, anglediff
from ..simulator import ParallelCalculated
from ..math.geometry import Geometry

__all__ = ['find_radius_of_photon_shell',
           'design_tilted_torus',
           'double_rowland_from_channel_distance',
           'add_offset_double_rowland_channels',
           'RowlandTorus',
           'ElementsOnTorus',
           'GratingArrayStructure',
           'RectangularGrid', 'CircularMeshGrid',
           ]


def find_radius_of_photon_shell(photons, mirror_shell, x, percentile=[1, 99]):
    '''Find the radius the photons coming from a single mirror shell have.

    For nested Wolter Type I mirrors the ray of photons reflected from a single
    mirror shell essentially form a cone in space. The tip of the cone is at
    the focal point and the base is at the mirror. There is a certain thickness
    to this cone depending on where exactly on the mirror the individual photon
    was reflected.

    This function takes a photon list of photons after passing through the
    mirror and calculates the radius range that this photon cone covers at a
    specific distance from the focal point. This information can help to design
    the placement of gratings.

    Parameters
    ----------
    photons : `~astropy.table.Table`
        Photon list with position and direction of photons leaving the mirror
    mirror_shell : int
        Select mirror shell to look at (uses column ``mirror_shell`` in
        ``photons`` for filtering).
    x : float
        Distance along the optical axis (assumed to coincide with the x axis
        with focal point at 0).
    percentile : list of floats
        The radius is calculated at the given percentiles. ``50`` would give
        the median radius. The default of ``[1, 99]`` gives a radius range
        excluding extreme outliers such as stray rays scattered into the extreme
        wing of the PSF.

    '''
    p = photons.copy()
    mdet = FlatDetector(position=np.array([x, 0, 0]), zoom=1e8, pixsize=1.)
    p = mdet(p)
    ind = (p['probability'] > 0) & (p['mirror_shell'] == mirror_shell)
    r = np.sqrt(p['det_x'][ind]**2+p['det_y'][ind]**2.)
    return np.percentile(r, percentile)


class RowlandTorus(MarxsElement, Geometry):
    '''Torus with y axis as symmetry axis.

    Note that the origin of the torus is the focal point, which
    **may or may not** be the same as the center of the torus.

    Parameters
    ----------
    R : float
        Radius of Rowland torus. ``r`` determines the radius of the Rowland
        circle, ``R`` is then used to rotate that circle around the axis of
        symmetry of the torus.
    r : float
        Radius of Rowland circle
    '''

    display = {'color': (1., 0.3, 0.3),
               'opacity': 0.2,
               'shape': 'torus; surface',
               'coo1': np.linspace(0, 2 * np.pi, 60),
               'coo2': np.linspace(0, 2 * np.pi, 60)}

    def __init__(self, R, r, **kwargs):
        self.R = R
        self.r = r
        # super does not work, because the various classes have different signature
        Geometry.__init__(self, kwargs)
        MarxsElement.__init__(self, **kwargs)

    def quartic(self, xyz, transform=True):
        '''Quartic torus equation.

        Roots of this equation are points on the torus.

        Parameters
        ----------
        xyz : np.array of shape (N, 3) or (3)
            Coordinates of points in euclidean space. The quartic is calculated
            for those points.
        transform : bool
            If ``True`` transform ``xyz`` from the global coordinate system
            into the local coordinate system of the torus. If this
            transformation is done in the calling function already, set to
        ``False``.

        Returns
        -------
        q : np.array of shape (N) or scalar
            Quartic at the input location
        '''
        if xyz.shape[-1] != 3:
            raise ValueError('Input coordinates must be defined in Euclidean space.')

        if transform:
            invpos4d = np.linalg.inv(self.pos4d)
            xyz = h2e(np.einsum('...ij,...j', invpos4d, e2h(xyz, 1)))
        return ((xyz**2).sum(axis=-1) + self.R**2. - self.r**2.)**2. - 4. * self.R**2. * (xyz[..., [0,2]]**2).sum(axis=-1)

    def solve_quartic(self, origin, v, transform=True):
        '''Solve the quartic on the Rowland torus.

        This method solves the quartic equation for positions on the Rowland
        Torus, i.e. it intersects a line with the torus. To that end, the
        location on the line is varied and the root of the quartic is found
        through numerical optimization.

        Parameters
        ----------
        origin : np.array
            Origin of line as homogeneous coordinate.
            This is also used as approximate starting point for the numerical
            optimization, so it should be reasonably close to the
            solution.
        v : np.array
            Direction of the line as homogeneous coordinate.
        transform : bool
            If ``True`` transform input from the global coordinate system
            into the local coordinate system of the torus. If this
            transformation is done in the calling function already, set to
            ``False``.

        Returns
        -------
        coo : np.array
            Position of intersection as homogeneous coordinate.

        '''
        origin = h2e(origin)
        v = h2e(v)
        def fun(k, origin, v):
            out = self.quartic(origin + k[..., None] * v, transform=transform)
            return out
        out = optimize.root(fun,
                            # The starting guess also serves to set the scale
                            # in the absence of a jacobian.
                            # So, can't start at 0. Instead, need to guess
                            # reasonable scale from torus parameters.
                            # r-R seems like a good guess but does not work at r = R.
                            # In that case, just pick a number that's small
                            # relative to r.
                            max(np.abs(self.r - self.R), self.r * 1e-2) * 0.1,
                            args=(origin, v))
        if not out.success:
            raise Exception('Intersection with torus not found.')
        else:
            return e2h(origin + out.x[..., None] * v, 1)

    def parametric_surface(self, theta, phi, display):
        '''Parametric representation of surface of torus.

        In contrast to `parametric` the input parameters here are 1-d arrays
        and the functions converts that to a rectangular grid.

        Parameters
        ----------
        theta : np.array
            Points on the Rowland circle (specified in rad, where 0 is on the
            optical axis across from the focal point).
        phi : np.array
            ``phi`` rotates the Rowland circle to make up the Rowland torus.

        Returns
        -------
        xyzw : np.array
            Torus coordinates in global homogeneous coordinate system.
        '''
        if (phi.ndim != 1) or (theta.ndim != 1):
            raise ValueError('input parameters have 1-dim shape.')
        theta, phi = np.meshgrid(theta, phi)
        return self.parametric(theta, phi)

    def parametric(self, theta, phi):
        '''Parametric description of points on the torus.

        This is just another way to obtain the shape of the torus, e.g.
        for visualization.

        Parameters
        ----------
        theta : np.array
            Points on the Rowland circle (specified in rad, where 0 is on the
            optical axis across from the focal point).
        phi : np.array
            ``phi`` rotates the Rowland circle to make up the Rowland torus.

        Returns
        -------
        xyzw : np.array
            Torus coordinates in global homogeneous coordinate system.
        '''
        x = (self.R + self.r * np.cos(theta)) * np.cos(phi)
        z = (self.R + self.r * np.cos(theta)) * np.sin(phi)
        y = self.r * np.sin(theta)
        w = np.ones_like(z)
        torus = np.array([x, y, z, w]).T
        return np.einsum('...ij,...j', self.pos4d, torus)

    def xyzw2parametric(self, xyzw, transform=True, intersectvalid=True):
        '''Calculate (theta, phi) coordinates for point on torus surface.

        Parameters
        ----------
        xyzw : np.array of shape (N, 4)
            Coordinates of points on the torus surface in homogeneous
            coordinates.

        transform : bool
            If ``True`` transform ``xyz`` from the global coordinate system
            into the local coordinate system of the torus. If this
            transformation is done in the calling function already, set to
            ``False``.

        intersectvalid : bool
            When ``r >=R`` the torus can intersect with itself. At these points,
            phi is not unique. If ``intersectvalid`` is true, those points will
            be filled with an arbitrarily chosen valid value (0.), otherwise
            they will be nan.

        Returns
        -------
        theta, phi: np.array
            Parametric representation for all points on the torus.
        '''
        if transform:
            invpos4d = np.linalg.inv(self.pos4d)
            xyzw = np.einsum('...ij,...j', invpos4d, xyzw)

        xyz = h2e(xyzw)

        if not np.allclose(self.quartic(xyz, transform=False) / self.R**4., 0.):
            raise ValueError('Parametric representation is only defined for points on torus surface.')
        # There are two branches and we can use the sign to distinguish between them
        s = np.sign(np.sqrt(xyz[:,0]**2 + xyz[:, 2]**2) - self.R)
        theta = np.arcsin(s * xyz[:, 1] / self.r) + (s < 0) * np.pi

        factor = self.R + self.r * np.cos(theta)
        phi = np.zeros_like(factor)
        ind = factor != 0
        phi[ind] = np.arctan2(xyz[ind, 2] / factor[ind], xyz[ind, 0] / factor[ind])
        if not intersectvalid:
            phi[~factor] = np.nan

        return theta, phi

    def normal_parametric(self, theta, phi):
        '''Return the gradient vector field.

        Following the usual convention, the vector is pointing outwards
        of the torus volume.

        Parameters
        ----------
        theta, phi : 1-d np.arrays
            Theta and phi coordinates for points on the torus surface.

        Returns
        -------
        gradient : np.array
            Gradient vector field in homogeneous coordinates. One vector
            corresponds to each input point. The shape of ``gradient`` is the
            same as the shape of ``xyz``.
        '''
        if not ((theta.ndim == 1) and (phi.ndim == 1)):
            raise ValueError('theta and phi must be 1-d arrays.')
        normal = np.empty((len(theta), 3))
        normal[:, 0] = np.cos(theta) * np.cos(phi)
        normal[:, 1] = np.sin(theta)
        normal[:, 2] = np.cos(theta) * np.sin(phi)

        return np.einsum('...ij,...j', self.pos4d, e2h(normal, 0))

    def normal(self, xyzw):
        '''Return the gradient vector field.

        Following the usual conventions, the vector is pointing outwards
        of the torus volume.

        Parameters
        ----------
        xyzw : np.array of shape (N, 4)
            Coordinates of points in euclidean space. The quartic is calculated
            for those points. All points need to be on the surface of the torus.

        Returns
        -------
        gradient : np.array
            Gradient vector field in euclidean coordinates. One vector
            corresponds to each input point. The shape of ``gradient`` is the
            same as the shape of ``xyz``.
        '''
        if not len(xyzw.shape) == 2:
            raise ValueError('Shape of input array must be (N, 4).')
        theta, phi = self.xyzw2parametric(xyzw, 1)
        return self.normal_parametric(theta, phi)

    def xyz_from_radiusangle(self, radius, angle, start):
        '''Get Cartesian coordinates for radius, angle on the rowland circle.

        y, z are calculated from the radius and angle of polar coordinates in a
        plane; then x is determined from the condition that the point lies on
        the Rowland circle.  The plane is perpendicular to the optical axis
        that defines the Rowland circle.

        Parameters
        ----------
        radius, angle : float or np.array of shape (n,)
            Polar coordinates in a plane perpendicular to the optical axis
            (where the optical axis is parallel to the x-axis and goes through
            the origin of the `RowlandTorus`.
            ``angle=0`` coincides with the local y-axis.
        start : number
            Starting value for the search. The quartic can have up to four
            solutions because a line can intersect a torus in four points.
            The solution found thus depends on where the numerical root solver
            starts the search.

        Returns
        -------
        xyz : np.array of shape (n, 3)
            Euclidean coordinates in the global coordinate system.
        '''
        y = radius * np.cos(angle)
        z = radius * np.sin(angle)
        # Use np.mean for backwards compatibility.
        # This used to take a bracketing interval instead of a first guess.
        out = self.solve_quartic(np.array([np.mean(start), y, z, 1]),
                               np.array([1, 0, 0, 0]),
                               transform=False)
        return h2e(np.einsum('...ij,...j', self.pos4d, out))


def design_tilted_torus(f, alpha, beta):
    """Design a torus with specifications similar to Heilmann et al. 2010

    A `RowlandTorus` is fully specified with the parameters ``r``, ``R`` and
    ``pos4d``.  However, in practice, these numbers might be derived from other
    values.  This function calculates the parameters of a RowlandTorus, based
    on a different set of input values.

    Parameters
    ----------
    f : float
        distance between focal point and on-axis grating. Should be as large as
        possible given the limitations of the spacecraft to increase the
        resolution.
    alpha : float (in radian)
        angle between optical axis and the line (on-axis grating - center of
        Rowland circle).
        A typical value could be twice the blaze angle.
    beta : float (in radian)
        angle between optical axis and the line (on-axis grating - hinge),
        where the hinge is a point on the Rowland circle. The Rowland Torus
        will be constructed by rotating
        the Rowland Circle around the axis (focal point - hinge).
        The center of the Rowland Torus will be the point where the line
        (on-axis grating - center of Rowland circle) intersects the line
        (focal point - hinge).

    Returns
    -------
    R : float
        Radius of Rowland torus. ``r`` determines the radius of the Rowland
        circle, ``R`` is then used to rotate that circle around the axis of
        symmetry of the torus.
    r : float
        Radius of Rowland circle
    pos4d : np.array of shape (4, 4)

    Notes
    -----
    The geometry used here really needs to be explained in a figure.
    However, some notes to explain at least the meaning of the symbols on the
    code are in order:

    - Cat : position of on-axis CAT grating (where the Rowland circle intersects the on-axis beam)
    """
    r = f / (2. * np.cos(alpha))
    R = r * np.sin(np.pi / 2 - beta)
    orientation = transforms3d.axangles.axangle2mat([0, 0, -1], beta - alpha)
    x_Ct = f / 2 - R * np.cos(beta - alpha)
    y_Ct = f / 2 * np.tan(alpha) + R * np.sin(beta - alpha)
    pos4d = transforms3d.affines.compose([x_Ct, y_Ct, 0], orientation, np.ones(3))
    return R, r, pos4d


def double_rowland_from_channel_distance(d_BF, R, f):
    '''This prescription is borne out of the engineering needs where it turns out
    that the spacing between channels d_BF is actually an important parameter
    so it makes sense to use that here to parameterize the torus, too.

    f is the distance between a theoretical on-axis CAT grating and the focal
    point.  It's default is a little smaller than 12 m, to leave space to mount
    the SPOs.

    Parameters
    ----------
    d_BF : float
        Distance between the optical axes of two channels measured
        in the dispersion direction.
    R : float
        Radius of Rowland torus.
    f : float
        Distance between a theoretical on-axis CAT grating and the focal
        point.

    Returns
    -------
    geometry : dict
        Dictionary with the keys that hold the input parameters,
        derived dimensions of the geometry, and the constructed
        Rowland tori. Three Rowland tori are constructed: One for
        each channel called (rowland_central, rowland_central_m) and
        one for the detector (rowland_detector).

    '''
    d = 0.5 * d_BF
    r = 0.5 * np.sqrt(f**2 + d_BF**2)
    alpha = np.arctan2(d_BF, f)
    pos = [f - (r + R) * np.cos(alpha),
           (r + R) * np.sin(alpha),
           0]
    orient = [[np.cos(alpha), np.sin(alpha), 0],
              [-np.sin(alpha), np.cos(alpha), 0],
              [0., 0., 1]]


    posm = [f - (r + R) * np.cos(-alpha),
           (r + R) * np.sin(-alpha),
            0]

    orientm = [
        [np.cos(-alpha), np.sin(-alpha), 0],
        [-np.sin(-alpha), np.cos(-alpha), 0],
        [0.0, 0.0, 1],
    ]

    geometry = {'d_BF': d_BF,
                'd': d,
                'f': f,
                'rowland_central': RowlandTorus(R=R, r=r, position=pos,
                                                orientation=orient),
                'rowland_central_m': RowlandTorus(R=R, r=r, position=posm,
                                                  orientation=orientm)}

    geometry['pos_det_rowland'] = np.array([0, -d, 0, 1])
    geometry['shift_det_rowland'] = np.eye(4)
    geometry['shift_det_rowland'][:, 3] = geometry['pos_det_rowland']
    geometry['rowland_detector'] = RowlandTorus(R, r,
                                                pos4d=geometry['rowland_central'].pos4d)
    geometry['rowland_detector'].pos4d = np.dot(geometry['shift_det_rowland'],
                                                geometry['rowland_detector'].pos4d)
    geometry['pos_opt_ax'] = {'1': np.array([0, -d, 0, 1]),
                              '1m': np.array([0, d, 0, 1])}
    return geometry


def add_offset_double_rowland_channels(geometry,
                                       offsets={'1': [0, -2.5, -7.5],
                                                '1m': [0, -2.5, -2.5],
                                                '2': [0, 2.5, 2.5],
                                                '2m': [0, 2.5, 7.5],
                                                }):
    '''Generate Rowland tori for multiple spectral channels in a double-tori design.

    While this function can generate four separate channels (splitting the
    original and the original mirrored channel in two each).
    Rowland tori are added to the input geometry dictionary.

    Parameters
    ----------
    geometry : dict
        Dictionary with the keys from ` double_rowland_from_channel_distance`.
    offsets : dict
        Dictionary with the keys that hold the channel ids and the
        values that hold the offsets in the x, y, and z direction.
        The default value is set for a total of four channels, where each
        of the original and the original mirrored channel is split in two.
        However, the function can be used to generate any number of channels
        with the naming convention that channels 1, 2, 3, ... are generated
        from the original channel and channels 1m, 2m, 3m, ... are generated
        from the original mirrored channel.
    '''
    d = geometry['d']
    geometry['pos_opt_ax'] = {}
    for k in offsets:
        sign = 1 if 'm' in k else -1
        geometry['pos_opt_ax'][k] = np.array([0, sign * d, 0, 1])

    for k, v in offsets.items():
        geometry['pos_opt_ax'][k][:3] += v
    # Now offset that Rowland torus in a z axis by a few mm.
    # Shift is measured from point on symmetry plane.
    for channel in geometry['pos_opt_ax']:
        shift_matrix = np.eye(4)
        shift_matrix[:, 3] = geometry['pos_opt_ax'][channel][:]
        if channel in ['1', '2']:
            base_rowland = 'rowland_central'
        elif channel in ['1m', '2m']:
            base_rowland = 'rowland_central_m'
        namer = 'rowland_' + channel
        geometry[namer] = RowlandTorus(geometry[base_rowland].R,
                                       geometry[base_rowland].r,
                                       pos4d=geometry[base_rowland].pos4d)
        geometry[namer].pos4d = np.dot(shift_matrix,
                                       geometry[namer].pos4d)


class ElementsOnTorus(ParallelCalculated, OpticalElement):
    '''A collection of elements on a Rowland torus.

    When initialized, it places elements in the space available on the
    Rowland torus, most commonly, this class is used to place grating facets.

    After generation, individual facet positions can be adjusted by hand by
    editing the attributes `elem_pos` or `elem_uncertainty`. See
    `marxs.simulation.Parallel` for details.

    After any of the `elem_pos`, `elem_uncertainty` or
    `uncertainty` is changed, `generate_elements` needs to be
    called to regenerate the facets on the GAS.

    Parameters
    ----------
    rowland : `marxs.design.rowland.RowlandTorus`
        Torus on which the elements are placed.
    d_element : list of two floats
        Size of the edge of elements along the two (y and z in canonical marxs orientation)
        edges.
        ``d_element`` can be larger than the actual size of the silicon
        membrane or the active area of a CCD chip to accommodate a minimum
        thickness of the surrounding frame.
    guess_distance : float
        A ray can intersect a torus in up to four
        points. ``guess_distance`` specifies the starting distance for the numerical search for
        the intersection point to resolve this ambiguity.
        Default is the start close to outer rim of the torus.
    optimize_axis : np.array
        Homogeneous coordinate of the axis along which elements will be moved. This will
        usually coincide with the optical axis of the telescope.
        Default is the x-axis in the coordinate system of the torus.
    normal_spec : np.array or callable
        see `~marxs.simulator.simulator.ParallelCalculated` for details.
        Default for this class: Pointing **inward** into the Rowlandtorus,
        as one would do for an array of e.g. CCDs.
    parallel_spec : np.array or callable
        see `~marxs.simulator.simulator.ParallelCalculated` for details.
        Default for this class: Symmetry axis of the torus.
    '''
    def __init__(self, **kwargs):
        self.rowland = kwargs.pop('rowland')
        # Default to a slightly positive number
        self.guess_distance = kwargs.pop('guess_distance',
                                         self.rowland.r + self.rowland.R)
        self.d_element = kwargs.pop('d_element')
        self.optimize_axis = np.asanyarray(kwargs.pop('optimize_axis',
                                           self.rowland.pos4d @ np.array([1, 0, 0, 0])))
        self.id_col = kwargs.pop('id_col', 'facet')

        # Defaults changes from parent classes. Document! Or remove?
        if 'normal_spec' not in kwargs.keys():
            kwargs['normal_spec'] = lambda xyzw: -self.rowland.normal(xyzw)
        if 'parallel_spec' not in kwargs.keys():
            # If not given, make parallel to Rowland circle at origin.
            kwargs['parallel_spec'] = self.rowland.pos4d @ np.array([0, 1, 0, 0])
        kwargs['pos_spec'] = self.elempos

        super().__init__(**kwargs)

    def elemposyz(self):
        '''Specify the element position in the yz plane

        For the torus, the Rowland circle lies in the xy plane and the symmetry
        axis of the torus is the y-axis. In this class, the position of the
        elements is specified in the yz plane in the coordinates of the
        torus and the remaining coordinate.

        This function will be customized by derived classes.

        Returns
        -------
        ypos, zpos : np.array
            1D arrays of y and z positions for the elements distributed in 2D.
        '''
        raise NotImplementedError

    def elempos(self):
        '''Generate 3D positions of elements

        For each element, y and z coordinates in the torus coordinate system
        are generated from ``self.elemposyz`` and the elements are placed at
        x=0. Then, that initial coordinate is transformed into the global
        coordinate system and moved by
        ``self.guess_distance * self.optimize_axis``. The resulting position
        is the starting point for numerical optimization along
        ``self.optimize_axis`` to find the on-torus location.
        '''
        ypos, zpos = self.elemposyz()
        posyz = np.vstack([np.zeros_like(ypos), ypos, zpos])
        origin = np.einsum('...ij,...j', self.rowland.pos4d, e2h(posyz.T, 1)) + self.guess_distance * self.optimize_axis
        # solve_quartic is not vectorized (because scipy.optimize is not)
        # so need to pass in one element at a time.

        # now, origin is in the global coordinate system.
        # Because transform=True is the default in solve_quartic that works.
        return np.vstack([self.rowland.solve_quartic(o, self.optimize_axis) for o in origin])


class RectangularGrid(ElementsOnTorus):
    '''A collection of diffraction gratings on the Rowland torus.

    This class is similar to ``marxs.design.rowland.GratingArrayStructure`` but
    instead of placing elements on concentric circles, they are placed to fill
    a rectangular area.

    When initialized, it places elements in the space available on the
    Rowland circle, most commonly, this class is used to place grating facets.

    After generation, individual facet positions can be adjusted by hand by
    editing the attributes `elem_pos` or `elem_uncertainty`. See
    `marxs.simulation.Parallel` for details.

    After any of the `elem_pos`, `elem_uncertainty` or
    `uncertainty` is changed, `generate_elements` needs to be
    called to regenerate the facets on the GAS.

    Parameters
    ----------
    y_range, z_range: list of two floats
        Limits of the rectangular area where gratings are placed.
        To place only one element, make both limits the same, e.g.
        ``z_range=[5, 5]`` will place one element in each row in z
        centered on $z=5$.
        This is given in the coordinate system of the torus, i.e. y is the
        plane of the Rowland circle and both numbers are measured from the
        center of the Rowland torus.
        Default for z-range: [-1e-10, 1e-10]. This will make one element
        centered on 0 in z-range.

    '''
    def __init__(self, **kwargs):
        self.y_range = kwargs.pop('y_range')
        self.z_range = kwargs.pop('z_range', [-1e-10, 1e-10])

        super().__init__(**kwargs)

    def elemposyz(self):
        n_y = int(np.ceil((self.y_range[1] - self.y_range[0]) / self.d_element[0]))
        n_z = int(np.ceil((self.z_range[1] - self.z_range[0]) / self.d_element[1]))
        n_y = max(1, n_y)
        n_z = max(1, n_z)

        # n_y and n_z are rounded up, so they cover a slightly larger range than y/z_range
        width_y = n_y * self.d_element[0]
        width_z = n_z * self.d_element[1]

        ypos = np.arange(0.5 * (self.y_range[0] - width_y + self.y_range[1] + self.d_element[0]), self.y_range[1], self.d_element[0])
        zpos = np.arange(0.5 * (self.z_range[0] - width_z + self.z_range[1] + self.d_element[1]), self.z_range[1], self.d_element[1])

        ypos, zpos = np.meshgrid(ypos, zpos)

        return ypos.flatten(), zpos.flatten()


class CircularMeshGrid(ElementsOnTorus):
    '''A collection of diffraction gratings on the Rowland torus filling a circle.

    When initialized, it places elements in the space available on the
    Rowland circle, most commonly, this class is used to place grating facets.

    After generation, individual facet positions can be adjusted by hand by
    editing the attributes `elem_pos` or `elem_uncertainty`. See
    `marxs.simulation.Parallel` for details.

    After any of the `elem_pos`, `elem_uncertainty` or
    `uncertainty` is changed, `generate_elements` needs to be
    called to regenerate the facets on the GAS.

    Parameters
    ----------
    radius : list of two floats
        Inner and outer radius of the circle.
    '''

    def __init__(self, **kwargs):
        self.radius = kwargs.pop('radius')
        super().__init__(**kwargs)

    def elemposyz(self):
        n_y = int(np.ceil(2 * self.radius[1] / self.d_element[0]))
        y_width = n_y * self.d_element[0]
        y_pos = np.arange(- y_width / 2, self.radius[1], self.d_element[0])

        ypos = []
        zpos = []

        for y in y_pos:
            if np.abs(y) > self.radius[1]:
                # Outermost layer. Center might be outside outer radius
                z = np.array([0])
            else:
                z_outer = np.sqrt(self.radius[1]**2 - y**2)
                if np.abs(y) > self.radius[0]:
                    n_z = int(np.ceil(2 * z_outer / self.d_element[1]))
                    z_width = n_z * self.d_element[1]
                    z = np.arange(- z_width / 2, z_outer, self.d_element[1])
                else:
                    z_inner = np.sqrt(self.radius[0]**2 - y**2)
                    z_mid = 0.5 * (z_inner + z_outer)
                    n_z = int(np.ceil((z_outer - z_inner) / self.d_element[1]))
                    z_width = n_z * self.d_element[1]
                    z = np.arange(z_mid - z_width / 2, z_outer, self.d_element[1])
                    z = np.hstack([-z, z])

            ypos.extend([y] * len(z))
            zpos.extend(z)
        return ypos, zpos


class GratingArrayStructure(ElementsOnTorus):
    '''Collection of diffraction gratings on the Rowland Torus

    When a ``GratingArrayStructure`` (GAS) is initialized, it places elements
    in the space available on the Rowland circle, most commonly, this class is
    used to place grating facets.

    After generation, individual facet positions can be adjusted by hand by
    editing the attributes `elem_pos` or `elem_uncertainty`. See `Parallel` for
    details.

    After any of the `elem_pos`, `elem_uncertainty` or `uncertainty` is
    changed, `generate_elements` needs to be called to regenerate the facets on
    the GAS.

    Parameters
    ----------
    radius : list of 2 floats
        Inner and outer radius of the GAS as measured in the yz-plane from the
        origin.

    phi : float or list of 2 floats
        Bounding angles for a segment covered by the GSA. :math:`\phi=0` is on
        the positive y axis. The segment fills the space from ``phi1`` to
        ``phi2`` in the usual mathematical way (counterclockwise).  Angles are
        given in radian. Note that ``phi[1] < phi[0]`` is possible if the
        segment crosses the y axis.
        Alternatively, ``phi`` can just be a single number. In that case, there
        will be exactly one element per radius.

    '''
    def __init__(self, **kwargs):
        self.phi = kwargs.pop('phi', [0., 2*np.pi])
        self.radius = kwargs.pop('radius')
        super().__init__(**kwargs)

    def max_elements_on_radius(self, radius):
        '''Distribute elements on a radius.

        Parameters
        ----------
        radius : list of two floats
            inner and outer radius that should be covered by elements

        Returns
        -------
        n : int
            Number of elements needed to cover a given radius segment.
            Elements might reach beyond the radius limits if the difference between
            inner and outer radius is not an integer multiple of the element size.
        '''
        return int(np.ceil((radius[1] - radius[0]) / self.d_element[0]))

    def distribute_elements_on_radius(self):
        '''Distributes elements as evenly as possible along a radius.

        .. note::
           Unlike `distribute_elements_on_arc`, this function will have
           elements reaching beyond the limits of the radius, if the distance
           between inner and outer radius is not an integer multiple of the
           element size.

        Returns
        -------
        radii : np.ndarray
            Radii of the element *center* positions.

        '''
        radii = []
        for i in range(len(self.radius) // 2):
            radiusbracket = self.radius[2 * i: 2 * i + 2]
            n = self.max_elements_on_radius(radiusbracket)
            radii.append(np.mean(radiusbracket) +
                         np.arange(- n / 2 + 0.5, n / 2 + 0.5) * self.d_element[0])
        return np.hstack(radii)


    def max_elements_on_arc(self, radius):
        '''Calculate maximal number of elements that can be placed at a certain radius.

        Parameters
        ----------
        radius : float
            Radius of circle where the centers of all elements will be placed.
        '''
        return radius * anglediff(self.phi) // self.d_element[1]

    def distribute_elements_on_arc(self, radius):
        '''Distribute elements on an arc.

        The elements are distributed as evenly as possible over the arc.

        .. note::

          Contrary to `distribute_elements_on_radius`, elements never stretch
          beyond the limits set by the ``phi`` parameter of the GAS. If an arc
          segment is not wide enough to accommodate at least a single element,
          it will go empty.
          (The exception to that is is ``phi`` is a single value. In that case,
          there will be exactly one element.)

        Parameters
        ----------
        radius : float
            radius of arc where the elements are to be distributed.

        Returns
        -------
        centerangles : array
            The phi angles for centers of the elements at ``radius``.

        '''
        if len(self.phi) == 1:
            return self.phi
        # arc is most crowded on inner radius
        n = self.max_elements_on_arc(radius - self.d_element[0] / 2)
        element_angle = self.d_element[1] / (2. * np.pi * radius)
        # thickness of space between elements, distributed equally
        d_between = (anglediff(self.phi) - n * element_angle) / (n + 1)
        centerangles = d_between + 0.5 * element_angle + np.arange(n) * (d_between + element_angle)
        return (self.phi[0] + centerangles) % (2. * np.pi)

    def elemposyz(self):
        angles = []
        radii = self.distribute_elements_on_radius()
        for r in radii:
            angles.append(self.distribute_elements_on_arc(r))
        # Turn lists of lists into flat numpy array
        radii = np.concatenate([[radii[i]] * len(a) for i, a in enumerate(angles)])
        angles = np.concatenate(angles)
        return radii * np.sin(angles), radii * np.cos(angles)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np


__all__ = ['rmid_spo', 'zoom_to_cover_all_spos']

def rmid_spo(conf):
    '''Looking at SPO positions, find midpoint of all SPOs.

    To simplify this code it only considers the mid-points of the SPOs, not
    their rotation angle. Since the rectangles are rotated a bit, e.g.
    the aperture needs to be a bit bigger than the area covered by SPOs.
    Also, if elements are above or below the SPOs, the beam may converge
    or come from the side. `azscale` and `rscale` can be set to adjust
    the outer corners a bit.

    Parameters
    ----------
    conf : dict
        Configuration dictionary with entry 'spo_geom' and 'spo_pos4d'

    Returns
    -------
    zoom : float
        middle of the SPO array in y direction
    '''
    spopos = np.array(conf['spo_pos4d'])
    return 0.5 * (spopos[:, 1, 3].max() + spopos[:, 1, 3].min())


def zoom_to_cover_all_spos(conf, azscale=.8, rscale=1.1):
    '''Looking at SPO positions, find zoom of rectangle that covers it.

    To simplify this code it only considers the mid-points of the SPOs, not
    their rotation angle. Since the rectangles are rotated a bit, e.g.
    the aperture needs to be a bit bigger than the area covered by SPOs.
    Also, if elements are above or below the SPOs, the beam may converge
    or come from the side. `azscale` and `rscale` can be set to adjust
    the outer corners a bit.

    Parameters
    ----------
    conf : dict
        Configuration dictionary with entry 'spo_geom' and 'spo_pos4d'
    azscale : float
        The outer edge is set to the center of the outermost SPO, plus
        `azscale` times the dimension in azimuth (global x coordinate)
    rscale : float
        The outer edge is set to the center of the outermost SPO, plus
        `rscale` times the dimension in radius (global y coordinate)

    Returns
    -------
    zoom : np.array of shape (3,)
        Zoom array as input to make pos4d
    '''
    spopos = np.array(conf['spo_pos4d'])
    azmid = 0.5 * (spopos[:, 0, 3].max() + spopos[:, 0, 3].min())
    rmid = rmid_spo(conf)
    delta_az = conf['spo_geom']['azwidth']
    delta_r = conf['spo_geom']['outer_radius'] - conf['spo_geom']['inner_radius']
    return np.array([1,
                     spopos[:, 0, 3].max() - azmid + azscale * delta_az.max(),
                     spopos[:, 1, 3].max() - rmid + rscale * delta_r.max()])


# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest

from marxs.missions.mitsnl.catgrating import NonParallelCATGrating
from marxs.missions.mitsnl.tolerances import wiggle_and_bar_tilt

from marxs.design.tests.test_tolerancing import gsa, elempos
from marxs.simulator import SimulationSetupError


def test_change_parallel_elements():
    '''Check that parameters work and elements are in fact changed.
    More detailed checks that the type of change is correct are
    implemented as separate tests, but those tests don't check out
    every parameter.
    '''
    g = gsa(NonParallelCATGrating)
    wiggle_and_bar_tilt(g, 0., 0., 0.)
    assert np.all(np.stack([e.pos4d for e in g.elements]) == elempos)

    for key in ['dx', 'dy', 'dz', 'rx', 'ry', 'rz']:
        d = {key: 1.23}
        wiggle_and_bar_tilt(g, **d)
        assert not np.all(np.stack([e.pos4d for e in g.elements]) == elempos)


def test_wiggle():
    '''Check wiggle function'''
    g = gsa(NonParallelCATGrating)
    wiggle_and_bar_tilt(g, dx=10, dy=.1)
    diff = elempos - np.stack([e.pos4d for e in g.elements])
    # Given the numbers, wiggle in x must be larger than y
    # This also tests that not all diff numbers are the same
    # (as they would be with move).
    assert np.std(diff[:, 0, 3]) > np.std(diff[:, 1, 3])


def test_wiggle_bar_tilt_error():
    '''Check that an error is raised if called on unsuitable object'''
    g = gsa()
    with pytest.raises(SimulationSetupError,
                       match='This wiggle function requires one NonParallelCATGrating per element.'):
        wiggle_and_bar_tilt(g)


def test_wiggle_bar_tilt():
    '''Check wiggle function. Here, we do not wiggle the elements, but just change
    the bar tilt and check that the resulting elements are moved.
    That's not to say they are moved correctly, but at least they are moved.'''
    g = gsa(NonParallelCATGrating)
    wiggle_and_bar_tilt(g, max_bar_tilt=.1)
    diff = elempos - np.stack([e.pos4d for e in g.elements])
    # Given the numbers, wiggle in x must be larger than y
    # This also tests that not all diff numbers are the same
    # (as they would be with move).
    assert np.all(diff[:, 1, :] == pytest.approx(0))
    assert np.all(diff[:, 3, :] == pytest.approx(0))
    # Rotation happens around these axes!
    assert np.all(diff[:, 0, :3] != 0)
    assert np.all(diff[:, 2, :3] != 0)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from copy import deepcopy
from scipy.interpolate import interp1d, RectBivariateSpline
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename

from marxs import optics
from marxs.simulator import Sequence
from marxs.base import SimulationSequenceElement

class MetaShellAperture(optics.MultiAperture):

    def __init__(self, conf, **kwargs):
        metashellgeometry = conf['metashellgeometry']
        elements = [optics.CircleAperture(position=[conf['focallength'] + 200, 0, 0],
                                          zoom=[1, shell['r_outer'], shell['r_outer']],
                                          r_inner=shell['r_inner'],
                                          id_num=shell['Metashell Serial Number'])
                    for shell in metashellgeometry]
        kwargs['elements'] = elements
        kwargs['id_col'] = 'shell'
        super().__init__(**kwargs)
        disp = {'color': (0.0, 0.75, 0.75),
                'opacity': 0.3,
                'shape': 'triangulation',
                'outer_factor': 1.3,
                'inner_factor': 1.}
        for i in range(len(self.elements) - 1):
            self.elements[i].display = deepcopy(disp)
            self.elements[i].display['outer_factor'] = 1.0 * metashellgeometry['r_inner'][i + 1] / metashellgeometry['r_outer'][i]
        self.elements[0].display['inner_factor'] = 0
        self.elements[-1].display = deepcopy(disp)
        self.elements[-1].display['outer_factor'] = 1.5


class MetaShellMirror(optics.FlatStack):

    display = {'shape': 'None'}

    def __init__(self, conf):
        metashellgeometry = conf['metashellgeometry']
        kwargs = {}
        kwargs['position'] = [conf['focallength'], 0, 0]
        kwargs['zoom'] = [20, np.max(metashellgeometry['r_outer'] * 1.1),
                          np.max(metashellgeometry['r_outer'] * 1.1)]
        kwargs['elements'] = [optics.PerfectLens,
                              optics.RadialMirrorScatter]
        kwargs['keywords'] = [{'focallength': conf['focallength']},
                              {'inplanescatter': conf['inplanescatter'],
                               'perpplanescatter': conf['perpplanescatter']}]

        super().__init__(**kwargs)


class MetaShellEfficiency(SimulationSequenceElement):
    def __init__(self, conf):
        data = conf['metashellgeometry']
        eff = conf['metashelleff']
        aeff = np.stack([eff[str(c)].data for c in data['Metashell Serial Number']])
        # factor 100 is to transform cm^2 to mm^2
        ageom = np.pi * (data['r_outer']**2 - data['r_inner']**2)
        self.relative_eff = aeff * 100 / ageom[:, None]
        self.shells = [(i, interp1d(eff['energy'], self.relative_eff[i - 1, :]))
                       for i in data['Metashell Serial Number']]

    def __call__(self, photons):
        for i, func in self.shells:
            ind = photons['shell'] == i
            prob = func(photons['energy'][ind])
            # Risk of extrapolation
            prob = np.clip(prob, 0., 1.)
            photons['probability'][ind] *= prob
        return photons


class MetaShell(Sequence):
    # pass all kwargs to MetaShellMirror.
    # This could be improved (e.g. to allow post_process_steps)
    # but not needed right now.
    def __init__(self, conf):
        super().__init__(elements=[MetaShellMirror(conf),
                                   MetaShellEfficiency(conf)])


class PerfectLensSegment(optics.PerfectLens):
    def __init__(self, conf, **kwargs):
        super().__init__(focallength=conf["focallength"], **kwargs)
        refl =  Table.read(get_pkg_data_filename(f"data/{conf['mirrorcoating']}"),
                           format='ascii.ecsv')
        self.reflectivity_interpolator = RectBivariateSpline(
            refl["energy"],
            [np.deg2rad(float(c)) for c in refl.colnames[1:]],
            np.stack([refl[c] for c in refl.columns[1:]]).T,
        )



# Licensed under GPL version 3 - see LICENSE.rst
from collections import OrderedDict
import inspect
from copy import deepcopy
import re
from datetime import datetime

import numpy as np
from transforms3d import affines
from astropy.table import Column

from ..visualization.utils import DisplayDict
from marxs import __version__

__all__ = ['GeometryError',
           'DocMeta',
           'MarxsElement',
           'TagVersion',
           'check_meta_consistent',
           'check_energy_consistent',
           'SimulationSequenceElement',
           ]


class GeometryError(Exception):
    pass


class DocMeta(type):
    '''Metaclass to inherit docstrings when required.

    When a derived class overwrites a method that was already defined in its
    base class, the new method usually has the same purpose as the original
    method and often uses the same parameters, too, although the implementation
    differs slightly.  In this case, it should have the same docstring, too.
    This metaclass will look for methods that are undocumented and add the
    docstring of the appropriate parent method to them.

    '''
    def __new__(mcs, name, bases, dict):
        # make a class here with the same method resolution order
        # but no attributes of its own (so that we can make it here without an
        # infinite loop, because making it will also go through this metaclass)
        if name == 'temporaryclass':
            return type.__new__(mcs, name, bases, dict)
        temp = type('temporaryclass', bases, {})
        mro = inspect.getmro(temp)

        for k in dict:
            # for items that have a docstring (i.e. methods) that is empty
            if hasattr(dict[k], '__doc__') and dict[k].__doc__ is None:
                # Some __doc__ strings cannot be rewitten in Python2, e.g. doc of a type
                try:
                    for b in mro:
                        # look if this b defines the method
                        # (it might not in case of multiple inheritance)
                        if hasattr(b, k) and (getattr(b, k).__doc__ is not None):
                            # copy docstring if super method has one
                            dict[k].__doc__ = getattr(b, k).__doc__
                            break
                except AttributeError:
                    pass

        return type.__new__(mcs, name, bases, dict)


class MarxsElement(metaclass=DocMeta):
    '''Base class for all elements in a MARXS simulation.

    This includes elements that actually change photon properties such
    as grating and mirrors, but also abstract concepts that do not
    have a direct hardware representation such as a "Rowland Torus".
    '''

    display = {'shape': 'None'}
    'Dictionary for display specifications, e.g. color'

    def __init__(self, **kwargs):
        '''Define a new MARXS element.'''
        if 'name' in kwargs:
            self.name = kwargs.pop('name')
        else:
            self.name = self.__class__

        if len(kwargs) > 0:
            raise ValueError('Initialization arguments {0} not understood'.format(', '.join(kwargs.keys())))
        self.display = DisplayDict(self, deepcopy(self.display))

    def describe(self):
        return OrderedDict(element=self.name)


reexp = re.compile(r"(?P<version>[\d.dev]+[\d]+)[+]?(g(?P<gittag>\w+))?[.]?(d(?P<dirtydate>[\d]+))?")
'''Regex to parse scm version string'''

ver = reexp.match(__version__)
'''Parsed version of the marxs code'''


class TagVersion(MarxsElement):
    '''Tag a photons list with diagnostic information such as a the program version.

    All keyword arguments passed when this element is initialized or when it is called
    will be added to the meta information of the photon list, some additional
    information on program version and runtime is added automatically.
    As such, the format of the keyword values is very flexible. However, if they are
    going to be written to fits files it is useful to follow fits conventions and
    the default paraemter values also invoke fits conventions.


    Parameters
    ----------
    origin : tuple of strings
        according to fits convention, the Institution where file was created
    creator : string or tuple of strings
        according to fits convention, the Person or program creating file'
    '''
    def __init__(self,
                 ORIGIN=('unkwown', 'Institution where file was created'),
                 CREATOR=('MARXS', 'Person or program creating file'),
                 **kwargs):
        super().__init__(name=kwargs.pop('name', self.__class__))

        kwargs['MARXSVER'] = (ver.group('version'), 'MARXS code version')
        if not ver.group('gittag') is None:
            kwargs['MARXSGIT'] = (ver.group('gittag'),
                                        'Git hash of MARXS code')
        if not ver.group('dirtydate') is None:
            kwargs['MARXSTIM'] = (ver.group('dirtydate'),
                                        'Date of dirty version ARCUS code')
        self.tags = kwargs

    def __call__(self, photons, *args, **kwargs):
        photons.meta['DATE'] = (datetime.now().isoformat()[:10], 'Date/time of computation')
        for k, v in self.tags.items():
            photons.meta[k] = v
        for k, v in kwargs.items():
            photons.meta[k] = v
        return photons


def check_meta_consistent(meta1, meta2, keywords=['ORIGIN', 'CREATOR',
                                                  'MARXSVER', 'MARXSGIT', 'MARXSTIM',
                                                  ],
                          allow_missing=True):
    '''Check that the meta data between two simulations indicates consistency.

    This check compares a number of keywords (e.g. the version of marxs) to
    indicate if the simulations where run with the same version of marxs.
    If a simulation records more information in the metadata,
    the check can be more complete.
    This function raises an `AssertionError` if the two sets of
    meta information are different.

    Parameters
    ----------
    meta1 : dict
        header from photon list 1
    meta2 : dict
        header from photon list 2
    keywords : list
        Keywords to be checked
    allow_missing : bool
        This flag allows a keyword to be missing from both dictionaries
        (e.g. MARXSGIT is not set if run with a released version); however,
        it is still an error if a key is present only in one, but not the
        dictionary.

    Raises
    ------
    AssertionError, KeyError
    '''
    for k in keywords:
        if (k in meta1) and (k in meta2):
            assert meta1[k] == meta2[k]
        else:
            if not allow_missing:
                raise KeyError(f'{k} not found in both dicts.')
            # Even with allow_missing=True
            # it's an error to have a key in only one dict
            if (k in meta1) or (k in meta2):
                raise KeyError(f'{k} found in one, but not both dicts.')


def check_energy_consistent(photons):
    '''Check if the energy is the same for all photons.

    If there is no energy colum, then this also passes.
    '''
    if 'energy' in photons.colnames:
        assert np.allclose(photons['energy'], photons['energy'][0])


class SimulationSequenceElement(MarxsElement):
    '''Base class for all elements in a simulation that processes photons.'''

    output_columns = []
    '''This is a list of strings that names the output properties.

    This gives the names of the output properties from this optical
    element.  `process_photon` or `process_photons` are responsible
    for calculating the values of these properties. For example, for a
    mirror of nested shells one might set
    ``output_columns = ['mirror_shell']``
    to pass the information on which shell the interaction took place to the
    user.

    The following properties are always included in the output and do not need
    to be listed here:

        dir : `numpy.ndarray`
            4-d direction vector of ray in homogeneous coordinates
        pos : `numpy.ndarray`
            4-d position of last interaction pf the photons with any optical
            element in homogeneous coordinates. Together with ``dir`` this
            determines the equation of the ray.
        energy : float
            Photon energy in keV.
        polarization : float
            Polarization angle of the photons.
        probability : float
            Probability that the photon continues. Set to 0 if the photon is
            absorbed, to 1 if it passes the optical element and to number
            between 0 and 1 to express a probability that the photons passes.

    '''

    id_col = None
    '''String that names an id column for output.

    Set this to a string to add an automatic numbering to the output. This is
    especially useful if there are several identical optical components that
    are used in parallel, e.g. there are four identical CCDs. Setting ``id_col
    = "CCD_ID"`` and passing an ``id_num=1, 2, 3, 4`` keyword respectively to
    each CCD will add a column ``CCD_ID`` with a value of 1,2,3, or 4 for each
    photon hitting one of those CCDs.

    Currently, this will not work with all optical elements.

    '''

    def __init__(self, **kwargs):
        self.id_num = kwargs.pop('id_num', -9)
        # We want to use id_col as a class attribute, but overwrite it if given as a kwarg
        if 'id_col' in kwargs:
            self.id_col = kwargs.pop('id_col')
        super().__init__(**kwargs)

    def add_output_cols(self, photons, colnames=[]):
        '''Add output columns to the photon array.

        This function takes the column names that are added to ``photons`` from
        several sources:

        - `id_col` (if not ``None``)
        - `output_columns`
        - the ``colnames`` parameter.

        Parameters
        ----------
        photons : `astropy.table.Table`
            Table columns are added to.
        colnames : list of elements
            Each element can be a string (in this case a float column with
            initial value ``np.nan`` is added) or a dictionay of arguments
            for `astropy.table.column.Column`. If the dictionay has a keys
            "value" then the column will be initialized to that value.
            Column names to be added; in addition several object properties can
            be used to set the column names, see description above.
        '''
        for n in self.output_columns + colnames:
            if (n is not None) and (n not in photons.colnames):
                if not isinstance(n, dict):
                    n = {'name': n, 'value': np.nan}
                val = n.pop('value', None)
                newcol = Column(length=len(photons), **n)
                if val is not None:
                    newcol[:] = val
                photons.add_column(newcol)

        # We can call this recursively, because the column will be added
        # before reaching this line agian, so we avoid infinite recursion.
        if (self.id_col is not None) and (self.id_col not in photons.colnames):
            self.add_output_cols(photons, colnames=[{'name': self.id_col,
                                                     'dtype': int,
                                                     'value': -1}])

    def __call__(self, photons, *args, **kwargs):
        return self.process_photons(photons, *args, **kwargs)


def _parse_position_keywords(kwargs):
    '''Parse keywords to define position.

    If ``pos4d`` is given, use that, otherwise look for ``position``, ``zoom``
    and ``orientation``.

    Parameters
    ----------
    pos4d : 4x4 array
        Transformation to bring an element from the default position
        (see description of individual elements) to a certain point in
        the space of homogeneous coordinates.
    position : 3-d vector in real space
        Measured from the origin of the spacecraft coordinate system.
    orientation : Rotation matrix or ``None``
        Relative orientation of the base vectors of this optical
        element relative to the orientation of the coordinate system
        of the spacecraft. The default is no rotation (i.e. the axes of both
        coordinate systems are parallel).
    zoom : float or 3-d vector
        Scale the size of an optical element in each dimension by
        ``zoom``. If ``zoom`` is a scalar, the same scale is applied
        in all dimesions. This only affects the oter dimensions of the
        optical element, not internal scales like the pixel size or
        grating constant (if defined for the optical element in
        question).  '''
    pos4d = kwargs.pop('pos4d', None)
    if pos4d is None:
        position = kwargs.pop('position', np.zeros(3))
        orientation = kwargs.pop('orientation', np.eye(3))
        zoom = kwargs.pop('zoom', 1.)
        if np.isscalar(zoom):
            zoom = np.ones(3) * zoom
        if not len(zoom) == 3:
            raise ValueError('zoom must have three elements for x,y,z or be a scalar (global zoom).')
        if np.any(np.array(zoom) <= 0):
            raise ValueError('All values in zoom must be positive-definite to keep pos4d matrix valid. Specify zero-thickness elements with display properties.')
        pos4d = affines.compose(position, orientation, zoom)
    else:
        if ('position' in kwargs) or ('orientation' in kwargs) or ('zoom' in kwargs):
            raise ValueError('If pos4d is specificed, the following keywords cannot be given at the same time: position, orientation, zoom.')

    if np.linalg.det(pos4d) == 0:
        raise ValueError('pos4d matrix is invalid (determinant is 0).')
    return pos4d



'''Provide simple defaults for a Arcus simulation

This module sets up a few simple defaults for ray-tracing things with Arcus.
It provides a source and pointing class for an on-axis pointing with some
sensibe defaults for energy, flux, jitter etc.
The purpose if this module is to save some typing when running simulations
from the command line and to make is easier to keep many different test
scripts in sync to all use the same parameters.
'''
from marxs.source import PointSource, JitterPointing
from marxs.math.utils import xyz2zxy
import astropy.units as u
from astropy.coordinates import SkyCoord
from marxs.missions.arcus.load_csv import load_number

coords = SkyCoord(30. * u.deg, 30. * u.deg)

sourcekwargs = {'coords': coords,
                'energy': 0.5 * u.keV,
                'flux': 1. / u.s / u.cm**2
                }
pointingkwargs = {'coords': coords,
                  'reference_transform': xyz2zxy,
                  }


class DefaultSource(PointSource):
    '''Default astronomical source for Arcus.

    This source will work with no parameters passed in and generate photons
    with a sensible pre-set for Arcus simulations (e.g. a fixed energy in
    the range that is interesting for Arcus).

    Parameters
    ----------
    Set any parameters that the base class would accept. Parameters are
    passed right through, the only difference in behavior to the base class is
    that defaults are provided for all required parameters (e.g. the
    coordinates on the sky).
    '''
    def __init__(self, **kwargs):
        for k in sourcekwargs:
            if k not in kwargs:
                kwargs[k] = sourcekwargs[k]
        super(DefaultSource, self).__init__(**kwargs)


class DefaultPointing(JitterPointing):
    '''Default astronomical source for Arcus.

    This pointing will work with no parameters passed in and it will set
    sensible defaults for an Arcus simulation (e.g. photons passing along
    the z-axis of the coordinate system for an on-axis source).

    Parameters
    ----------
    Set any parameters that the base class would accept. Parameters are
    passed right through, the only difference in behavior to the base class is
    that defaults are provided for all required parameters (e.g. the jitter and
    the pointing direction).
    '''

    def __init__(self, **kwargs):
        for k in pointingkwargs:
            if k not in kwargs:
                kwargs[k] = pointingkwargs[k]
        if 'jitter' not in kwargs:
            kwargs['jitter'] = load_number('other', 'pointingjitter',
                           'FWHM') / 2.3545
        super(DefaultPointing, self).__init__(**kwargs)



import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d

from marxs.source.labSource import LabPointSource
from marxs.optics.multiLayerMirror import MultiLayerMirror
from marxs.optics.detector import FlatDetector

'''This is an example test script that uses a source, multilayer mirror, and detector.
Note that string1 and string2 depend on the paths of those files.
'''

def normalEnergyFunc(mean, stdDev):
    '''This creates and returns a function that assigns energies to photons randomly
    according to the normal distribution with the given mean and standard deviation.
    '''
    def energyFunc(photonTimes):
        n = len(photonTimes)
        return np.random.normal(mean, stdDev, n)
    return energyFunc


string1 = '../marxs/optics/data/A12113.txt'
string2 = '../marxs/optics/data/ALSpolarization2.txt'
a = 2**(-0.5)
rotation1 = np.array([[a, 0, -a],
					  [0, 1, 0],
					  [a, 0, a]])
rotation2 = np.array([[0, 0, 1],
					  [0, 1, 0],
					  [-1, 0, 0]])

source = LabPointSource([100., 0., 0.], direction='-x', energy = normalEnergyFunc(0.31, 0.003), flux = 1e9)
mirror = MultiLayerMirror(string1, string2, position=np.array([0., 0., 0.]), orientation=rotation1)
detector = FlatDetector(1., position=np.array([0., 0., 25.]), orientation=rotation2, zoom = np.array([1, 100, 100]))

photons = source.generate_photons(0.01)
photons = mirror.process_photons(photons)
photons = detector.process_photons(photons)

p = photons[photons['probability'] > 0]
r = np.random.uniform(size=len(p))

q = p[p['probability'] > r]

plt.clf()
fig = plt.figure()
ax = fig.gca()
ax.plot(q['det_x'], q['det_y'], '.')
ax.set_xlim([-100,100])
ax.set_ylim([-100,100])
plt.show()

fig2 = plt.figure()
ax2 = fig2.gca(projection='3d')
#ax2.quiver(np.zeros(len(q)), np.zeros(len(q)), np.zeros(len(q)), q['polarization'][:,0], q['polarization'][:,1], q['polarization'][:,2])
ax2.quiver(q['pos'][:, 0] * 10, q['pos'][:, 1] * 10, q['pos'][:, 2] * 10, q['polarization'][:,0], q['polarization'][:,1], q['polarization'][:,2], length = 15)
ax2.set_xlim([-100,100])
ax2.set_ylim([-100,100])
ax2.set_zlim([-350,-150])
plt.show()

a = sum(abs(q['polarization'][:,0]))
b = sum(abs(q['polarization'][:,1]))

#print a,b


# Licensed under GPL version 3 - see LICENSE.rst
'''This module contains helper functions to make running simulations easier.

Sometimes, those functions save only a few lines of code, but that can still
improve the readability of e.g. loops and cells in notebooks.
'''
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as u
from marxs.source import PointSource, FixedPointing


__all__ = ['default_coord',
           'monoenergetic_astrosimulation',
           ]

default_coord = SkyCoord(30., 30., unit='deg')

def monoenergetic_astrosimulation(instrument, energy, n_photons=2e4,
                                  reference_transform=np.eye(4)):
    '''Run simple simulation at fixed energy and fixed pointing.

    Parameters
    ----------
    instrument : `marxs.simulator.simulator.Sequence`
        Definition of an instrument with an aperture, but no pointing.
        This function will add a simple fixed pointing and a point source.
    energy : quantity
        Energy for the simulation
    n_photon : int
        number of photons for the simulation
    reference_transform : np.array of shape (4, 4)
        For instruments that do not use the canonical x-axis as the optical axis.
        See `marxs.source.pointing.FixedPointing`


    Returns
    -------
    photons : `astropy.table.Table`
        photon list
    '''
    mysource = PointSource(coords=default_coord,
                           energy=energy,
                           flux=1. / u.s / u.cm**2)
    fixedpointing = FixedPointing(coords=default_coord, reference_transform=reference_transform)
    photons = mysource.generate_photons(n_photons * u.s)
    photons = fixedpointing(photons)
    photons = instrument(photons)
    return photons



# Licensed under GPL version 3 - see LICENSE.rst
import os
from os.path import join as pjoin
import string

import numpy as np
import astropy.units as u
from astropy.table import Table, QTable
from scipy.interpolate import interp1d

from marxs.missions.arcus.utils import config
from marxs.missions.arcus.arcus import defaultconf, DetCamera, FiltersAndQE
from marxs.missions.arcus.load_csv import load_table
from marxs.missions.arcus.utils import tagversion
from marxs.reduction import ogip


def filename_from_meta(filetype='fits', **kwargs):
    '''Generate default filename from ARF or RMF header values

    Parameters
    ----------
    filetype : string
        file ending (the file is usually saved as fits, but the ending
        might be "arf", "rmf", or any other string)
    kwargs: string/int
        Keywords used to construct filename
    '''
    chan = kwargs.pop('ARCCHAN')
    if chan == '1111':
        chan = 'all'  # special case for readability
    filename = f'chan_{chan}'
    return filename + ogip.filename_from_meta(filetype=filetype, **kwargs)


class OnCCD:
    '''Class that calculates if a wavelength falls on a CCD or not

    Some of the Arcus parameters are taken from the dictionary, others (such as
    the position of the CCDs are taken from loading in the Arcus simulator
    and inspecting its objects.

    Parameters
    ----------
    conf : dict
        Dictionary with Arcus design parameters
    '''
    def __init__(self, conf):
        self.conf = conf
        camera = DetCamera(conf)
        self.corners = np.zeros((len(camera.elements), 2, 2))

        for i, ccd in enumerate(camera.elements):
            self.corners[i, 0, :] = (ccd.geometry['center'] -
                                     ccd.geometry['v_y'])[[0, 2]]
            self.corners[i, 1, :] = (ccd.geometry['center'] +
                                     ccd.geometry['v_y'])[[0, 2]]
        facets = load_table('gratings', 'facets')
        self.effective_grating_period = np.cos(conf['blazeang'] * u.degree) * np.mean(facets['period']) * facets['period'].unit
        # without the blaze angle would look like this
        # effective_grating_period = np.mean(facets['period']) * \
        #    facets['period'].unit

    def __call__(self, channel_mid, order, opt_ax):
        '''Mirror and grating efficiency

        This function interpolates the mirror and grating efficiency on
        a given energy grid.

        Parameters
        ----------
        channel_mid : `~astropy.units.quantity.Quantity`
            Representative energy for each channel. Grating efficiency etc. are
            interpolated on this energy, so bins should be narrow enough to
            resolve any features.
        order : ind
            Diffraction order
        opt_ax : string
            Name of the optical axes
        conf : dict
            Arcus configuration

        Returns
        -------
         transmission : np.array
            Transmission in the range 0..1 for each bin.
         '''
        opt_ax_x = self.conf['pos_opt_ax'][opt_ax][0]
        # np.sign accounts for the fact that some orders are
        # reversed in direction
        alpha = np.arctan2(np.sign(opt_ax_x) *
                           (self.corners[:, :, 0] - opt_ax_x),
                           self.conf['f'] - self.corners[:, :, 1]) * u.rad
        theta = np.arcsin(order * u.dimensionless_unscaled *
                          channel_mid.to(u.Angstrom,
                                         equivalencies=u.spectral()) /
                          self.effective_grating_period)
        onanyccd = np.zeros(len(theta), dtype=bool)

        for row in alpha:
            ind = (theta > min(row)) & (theta < max(row))
            onanyccd[ind] = True
        return onanyccd


class FiltersQE(FiltersAndQE):
    def __call__(self, en_mid):
        # Make it look like a photon table so that I can call MARXS element
        tab = Table(data=[en_mid, np.ones(len(en_mid))],
                    names=['energy', 'probability'])
        tab = super().__call__(tab)
        return tab['probability']


class MirrGrat:
    def __init__(self, aefforder,
                 kind='quadratic', fill_value=0., bounds_error=False):
        # Bug in astropy <=4.2:
        # need to wrap into extra QTable or sort will fail
        aefforder = QTable(aefforder, copy=True)
        aefforder.sort("wave")
        self.interp = {}
        self.unit = {}
        for o in aefforder.colnames[1:]:
            self.interp[o] = interp1d(aefforder['wave'], aefforder[o],
                                      kind=kind, fill_value=fill_value,
                                      bounds_error=bounds_error)
            # interp1d does not conserve unit, so need to keep separately
            self.unit[o] = aefforder[o].unit

    def __call__(self, channel_mid, order):
        '''Mirror and grating efficiency

        This function interpolates the effective areas in the caldb.

        Parameters
        ----------
        channel_mid : `~astropy.units.quantity.Quantity`
            Representative energy for each channel. Grating efficiency etc. are
            interpolated on this energy, so bins should be narrow enough to
            resolve any features.
        order : ind
            Diffraction order

        Returns
        -------
        transmission : np.array
            Transmission in the range 0..1 for each bin.
        '''
        transmission = self.interp[str(order)](channel_mid.to(u.Angstrom,
                                               equivalencies=u.spectral()))

        # Just to prevent unphysical extrapolation
        return np.clip(transmission, 0, None) * self.unit[str(order)]


def mkarf(channel_edges, order,
          mirr_grat=None, trans_filters_qe=None,
          conf=defaultconf,
          channels=list(defaultconf['pos_opt_ax'].keys())):
    '''Make an ARF for 0th order or a grating spectrum

    Parameters
    ----------
    channel_edges : `~astropy.units.quantity.Quantity`
        Edges of the channels in energy or wavelength. The ARF that is
        generated from  n edges will have n-1 bins.
    order : int
        Diffraction order
    mirr_grat : callable or None
        Function that interpolates mirror and grating efficiency. If None,
        it is read from ``mirr_grat.tab`` in the caldb.
    trans_filters_qe : callable or None
        Function that interpolates filter curves and CCD QE. If None,
        it is initialized from `FiltersQE` which uses Arcus defaults and
        allows extrapolation.
    conf : dict
        Arcus configuration
    channels : list
        List of channel names for an ARF. ARFs can be made for a single order
        or be combined for a set of orders.

    Returns
    -------
    transmission : np.array
        Transmission in the range 0..1 for each bin.
    '''
    energy_edges = channel_edges.to(u.keV, equivalencies=u.spectral())
    energy_edges = np.sort(energy_edges)
    en_mid = 0.5 * (energy_edges[:-1] + energy_edges[1:])

    onccd = OnCCD(conf)

    if mirr_grat is None:
        aefforder = QTable.read(os.path.join(config['data']['caldb_inputdata'], 'aeff',
                                     'mirr_grat.tab'), format='ascii.ecsv')
        mirr_grat = MirrGrat(aefforder)

    if trans_filters_qe is None:
        trans_filters_qe = FiltersQE(kwargs_interp1d={'bounds_error': False,
                                       'fill_value': 0.})
    specresp_filtqe = trans_filters_qe(en_mid)
    specresp_sim = mirr_grat(en_mid, order)
    all_onccd = np.zeros(len(specresp_filtqe))
    for chan in channels:
        if chan not in defaultconf['pos_opt_ax'].keys():
            raise ValueError(f'channel {chan} not defined in Arcus')
        specresp_onccd = onccd(en_mid, order, chan)
        all_onccd += specresp_onccd

    arf = ogip.ARF(data=[energy_edges[:-1], energy_edges[1:],
                    specresp_sim * specresp_filtqe * all_onccd,
                    specresp_sim, specresp_filtqe, all_onccd,
                    energy_edges[1:].to(u.Angstrom,
                                        equivalencies=u.spectral()),
                    energy_edges[:-1].to(u.Angstrom,
                                         equivalencies=u.spectral()),
                    ],
              names=['ENERG_LO', 'ENERG_HI', 'SPECRESP', 'MIRROR_GRAT',
                     'FILTERS_QE', 'ONCCD',
                     # CIAO adds these to grating ARFs, so we do, too:
                     'BIN_LO', 'BIN_HI',
                     ])
    arf.meta['ARCCHAN'] = ''.join(np.isin(list(conf['pos_opt_ax'].keys()),
                                          channels).astype(int).astype(str))
    tagversion(arf, ORDER=order)
    return arf


## Can I replace this with ogiprmffrom_Gauss_sigma?
# I wrote it for that purpose, but need to compare output to make sure it works that same way.
def mkrmf0(bin_edges, *, ccdfwhm: QTable, threshold=1e-6):
    '''Make RMF for 0 order

    The RMF is a single Gaussian.

    Parameters
    ----------
    channel_edges : `~astropy.units.quantity.Quantity`
        Edges of the channels in energy or wavelength. The RMF that is
        generated from n edges will have n-1 bins.
    threshold : float
        To reduce the file size, RMF components below the threshold value
        are cut-off
    ccdfwhm : `astropy.table.Table`
        Table with CCD properties. Two columns are expected with
        'energy' (in keV) and 'FWHM' (in eV). Note that the units are
        hardcoded and different between columns.
    '''
    from sherpa.models import NormGauss1D

    ebounds = ogip.RMF.ebounds_from_edges(bin_edges)
    matrix = Table(names=['ENERG_LO', 'ENERG_HI', 'N_GRP',
                          'F_CHAN', 'N_CHAN', 'MATRIX'],
                   data=[np.zeros(len(ebounds), dtype=d) for d in [
                       np.float32, np.float32, np.int16,
                       object, object, object]],
                   units=[u.keV, u.keV, None, None, None, None])

    matrix['ENERG_LO'] = ebounds['E_MIN']
    matrix['ENERG_HI'] = ebounds['E_MAX']
    matrix.meta['DETCHANS'] = len(bin_edges) - 1
    matrix.meta['LO_THRES'] = threshold
    matrix.meta['TLMIN{}'.format(matrix.colnames.index('F_CHAN') + 1)] = 0
    ebounds.meta['TLMIN{}'.format(ebounds.colnames.index('CHANNEL') + 1)] = 0

    for tab in [ebounds, matrix]:
        tab = tagversion(tab, ORDER=0)

    for r, miden in enumerate(0.5 * (ebounds['E_MIN'] + ebounds['E_MAX'])):
        func = NormGauss1D('name')
        func.pos = miden
        func.FWHM = np.interp(miden, ccdfwhm['energy'],
                              ccdfwhm['FWHM'].to(u.keV))  # eV to keV
        fullmatrix = func(ebounds['E_MIN'], ebounds['E_MAX'])
        out = ogip.RMF.arr_to_rmf_matrix_row(fullmatrix, 0,
                                             threshold=threshold)
        for i, col in enumerate(['N_GRP', 'F_CHAN', 'N_CHAN', 'MATRIX']):
            matrix[col][r] = out[i]
    return ogip.RMF(matrix, ebounds, CHANTYPE='PHA', HDUCLAS3='REDIST')


def write_readme(osip, outpath, outroot='', ARCCHAN='all'):
        '''Write README file to directory with ARFs

        Parameters
        ----------
        osip : `marxs.reduction.osip.OSIPBase` object
            OSIP objet that generates data in this directory.
        outpath : string
            Location where the output ARFs are deposited
        outroot : string
            prefix for output filename
        '''
        # Get a table so that I can tag it and get all the meta information
        tag = Table()
        tagversion(tag)
        # Format the meta information into a string
        tagstring = ''
        for k, v in tag.meta.items():
            if isinstance(v, str):
                tagstring += f'{k}: {v}\n'
            else:
                tagstring += f'{k}: {v[0]}   // {v[1]}\n'

        with open(pjoin(os.path.dirname(__file__),
                        'data', "osip_template.md")) as t:
            template = string.Template(t.read())

        output = template.substitute(tagversion=tagstring,
                                     description=f'{osip.__class__}: {osip.osip_description}',
                                     offsetorders=osip.offset_orders,
                                     filenamemain=filename_from_meta(filetype='arf', ARCCHAN=ARCCHAN, ORDER=-4, CCDORDER=-4, TRUEORD=-4),
                                     filenameupper=filename_from_meta(filetype='arf', ARCCHAN=ARCCHAN, ORDER=-4, CCDORDER=-4, TRUEORD=-5),
                                     filenamelower=filename_from_meta(filetype='arf', ARCCHAN=ARCCHAN, ORDER=-4, CCDORDER=-4, TRUEORD=-3))
        with open(pjoin(outpath, "README.md"), "w") as f:
            f.write(output)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from ..ogip import RMF


def test_compress_rmf_matrix():
    arr = np.array([0., 1e-9, .1, .2, .1, 1e-3, .1])
    ngrp, fchan, nchan, mat = RMF.arr_to_rmf_matrix_row(arr, 1,
                                                        threshold=1e-2)
    assert ngrp == 2
    assert fchan == [3, 7]
    assert nchan == [3, 1]
    assert mat == [.1, .2, .1, .1]

    ngrp, fchan, nchan, mat = RMF.arr_to_rmf_matrix_row(arr, 0,
                                                        threshold=1e-2)
    assert ngrp == 2
    assert fchan == [2, 6]
    assert nchan == [3, 1]
    assert mat == [.1, .2, .1, .1]



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest
import astropy.units as u

from marxs.missions.arcus.arcus import (PerfectArcus,
                                        defaultconf as conf)
from marxs.missions.arcus.defaults import DefaultSource, DefaultPointing
from marxs.missions.arcus.utils import id_num_offset
from ..analyze_grid import zeropos

@pytest.mark.xfail(reason="Arcus configuration is changing, so this might fail.")
def test_layout_to_EdHertz():
    '''Compare derived torus tilt angle to number calculated by
    Ed Hertz and Casey.'''
    assert np.isclose(np.rad2deg(np.sin(conf['rowland_central'].pos4d[2, 1])), 2.89214, rtol=1e-3)

class DirectLight(PerfectArcus):
    '''Default Definition of Arcus without any misalignments'''
    gratings_class = None

pointing = DefaultPointing()
instrument = DirectLight()
source = DefaultSource(energy=1. * u.keV)
photons = source.generate_photons(40_000 * u.s)
photons = pointing(photons)
photons = instrument(photons)
n_apertures = len(list(id_num_offset.values()))
photons['aper_id'] = np.digitize(photons['xou'],
                                     bins=list(id_num_offset.values()))
chan_name = list(id_num_offset.keys())

@pytest.mark.parametrize('channel', ['1', '1m', '2', '2m'])
@pytest.mark.parametrize('col',['proj_x', 'proj_y', 'circ_phi', 'circ_y'])
def test_zeropos(channel, col):
    '''Test the zeropos function'''
    ind = photons['aper_id'] == chan_name.index(channel) + 1
    assert zeropos(col, channel, conf) == pytest.approx(np.mean(photons[col][ind]), rel=1e-2)

def test_arcus_zeropos():
    '''Apply the zeropos function to test the Arcus zero order positions'''
    assert np.abs(zeropos('proj_y', '1', conf) - zeropos('proj_y', '1m', conf)) > 4
    assert np.abs(zeropos('proj_y', '2', conf) - zeropos('proj_y', '1', conf)) > 4
    assert np.abs(zeropos('proj_y', '2', conf) - zeropos('proj_y', '2m', conf)) > 4



import numpy as np
from numpy.random import rand, randn

from marxs.missions.mitsnl.catgrating import InterpolateEfficiencyTable, NonParallelCATGrating
from marxs.design.uncertainties import generate_facet_uncertainty as genfacun
from marxs.design.tolerancing import oneormoreelements
from marxs.simulator import SimulationSetupError

__all__ = ['OrderSelectorWavy',
           'OrderSelectorTopHat',
           'wiggle_and_bar_tilt',
           ]

class OrderSelectorWavy(InterpolateEfficiencyTable):
    '''Add a random number to blaze angle before looking up Ralf Table

    In the lab, it seems that the grating bars are not exactly
    perpendicular to the "surface" of the grating. This class adds
    a random number drawn from a Gaussian distribution to the blaze angle
    before looking up the grating efficiency and selecting the order.

    Parameters
    ----------
    wavysigma : float
        Sigma of Gaussian distribution (in radian)
    '''
    def __init__(self, wavysigma, **kwargs):
        self.sigma = wavysigma
        super().__init__(**kwargs)

    def probabilities(self, energies, pol, blaze):
        return super().probabilities(energies, pol, blaze + self.sigma * randn(len(blaze)))


class OrderSelectorTopHat(InterpolateEfficiencyTable):
    '''Add a random number to blaze angle before looking up Ralf Table

    In the lab, it seems that the grating bars are not exactly
    perpendicular to the "surface" of the grating. This class adds
    a random number drawn from a Gaussian distribution to the blaze angle
    before looking up the grating efficiency and selecting the order.

    Parameters
    ----------
    tophatwidth : float
        width of tophat function (in radian)
    '''
    def __init__(self, tophatwidth, **kwargs):
        self.tophatwidth = tophatwidth
        super().__init__(**kwargs)

    def probabilities(self, energies, pol, blaze):
        return super().probabilities(energies, pol, blaze + self.tophatwidth * (rand(len(blaze)) - 0.5))


@oneormoreelements
def wiggle_and_bar_tilt(e, dx=0, dy=0, dz=0, rx=0., ry=0., rz=0., max_bar_tilt=0.):
    '''Move and rotate elements around principal axes accounting for bar tilt.

    This function is made for `~marxs.optics.CATGrating`, and simulates a situation
    where the bar angles are not normal to the grating surface.
    We draw a bar tilt angle for each grating from a given distribution. Then,
    an element is rotated to compensate this bar tilt angle. This represents the a
    correction that happens when the grating is bonded into a frame.
    Additionally, the "normal" random misalignment is applied.

    Parameters
    ----------
    e : `marxs.simulator.Parallel` or list of those elements
        Elements where uncertainties will be set
    dx, dy, dz : float
        accuracy of positioning in x, y, z (in mm) - Gaussian sigma, not FWHM!
    rx, ry, rz : float
        accuracy of positioning. Rotation around x, y, z (in rad) - Gaussian
        sigma, not FWHM!
    max_bar_tilt : float
        Maximum value for bar tilt, drawn from a uniform distribution.
    '''
    tilts = np.random.uniform(low=-max_bar_tilt, high=max_bar_tilt, size=len(e.elements))
    tilt_offsets = np.zeros((len(e.elements), 3))
    tilt_offsets[:, 1] = tilts
    e.elem_uncertainty = genfacun(len(e.elements), [dx, dy, dz], [rx, ry, rz],
                                  rot_offset=-tilt_offsets)
    e.generate_elements()

    # Find the CAT grating membrane in the stack
    # e.g. L1 and L2 are sub-classes, we only want to apply this to main grating
    grats = e.elements_of_class(NonParallelCATGrating, subclass_ok=False)
    if len(grats) != len(tilts):
        raise SimulationSetupError(f'Number of elements of {e} does not match number of NonParallelCATGratings. \n' +
                             'This wiggle function requires one NonParallelCATGrating per element.')

    for elem, t in zip(grats, tilts):
        if not hasattr(elem, 'blaze_center'):
            raise ValueError(f'Object {elem} does not have a blaze_center attribute.')

        elem.blaze_center = t



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from ..ogip import ColOrKeyTable

import pytest

@pytest.mark.parametrize('input', [3, [3, 3], 3 * np.ones(2, dtype=int)])
def test_insert_allsame(input):
    t = ColOrKeyTable([['text1', 'text2'], [1, 2]], names=['a', 'b'])
    t['c'] = input
    assert t.meta['c'] == 3
    assert t['c'] == 3
    assert t.colnames == ['a', 'b']


def test_access_meta_as_column():
    t = ColOrKeyTable([['text1', 'text2'], [1, 2]], names=['a', 'b'])
    t.meta['test'] = 'mytext'
    assert t['test'] == 'mytext'


def test_existing_col_allsame():
    t = ColOrKeyTable([['text1', 'text2'], [1, 2]], names=['a', 'b'])
    t['a'] = 'text3'
    assert t.meta['a'] == 'text3'
    assert t['a'] == 'text3'
    assert t.colnames == ['b']


def test_existing_col_allsame_index():
    t = ColOrKeyTable([['text1', 'text2'], [1, 2]], names=['a', 'b'])
    t['a'][1] = 'text1'
    assert t.meta['a'] == 'text1'
    assert t['a'] == 'text1'
    assert t.colnames == ['b']


def test_existing_col_allsame_indexarray():
    t = ColOrKeyTable([['text1', 'text2', 'text2'], [1, 2, 3]], names=['a', 'b'])
    t['a'][t['a'] == 'text2'] = 'text1'
    assert t.meta['a'] == 'text1'
    assert t['a'] == 'text1'
    assert t.colnames == ['b']

@pytest.mark.skip()
def test_setitem_on_meta_as_column():
    '''Have information that's in meta, but then change one element'''
    t = ColOrKeyTable([['text1', 'text2'], [1, 2]], names=['a', 'b'])
    t.meta['test'] = 'mytext'
    t['test'][1] = 'mytext_different'
    assert t.colnames == ['a', 'b', 'test']
    assert np.all(t['test'] == ['mytext', 'mytext', 'mytext_different'])


# Licensed under GPL version 3 - see LICENSE.rst
import pytest
import numpy as np
from astropy.table import Table

from ..base import (TagVersion, _parse_position_keywords,
                    check_meta_consistent, check_energy_consistent,
)
from ..base import SimulationSequenceElement as SSE


def test_invalid_zoom():
    '''This error is easy to make, e.g. that the thickness of
    an aperture to 0, thus is deserves its own test and its own
    error message.
    '''
    with pytest.raises(ValueError) as e:
        _parse_position_keywords({'zoom': [0, 0, 1]})
    assert 'All values in zoom must be positive' in str(e.value)


def test_invalid_pos4d():
    pos4d = np.array([[0,  0, -7.5, -1.35],
                      [0.,  5.5,  0.,  0],
                      [0.,  0.,  4.e-15,  1.4],
                      [0.,  0.,  0.,  1.]])
    with pytest.raises(ValueError) as e:
        _parse_position_keywords({'pos4d': pos4d})
    assert "is invalid" in str(e.value)


def test_add_output_cols():
    '''Test different ways to set up output columns'''
    class T(SSE):
        output_columns = ['a', 'b']

    photons = Table({'a': [1, 2]})
    t = T()
    t.add_output_cols(photons)
    assert photons.colnames == ['a', 'b']
    # An existing column is not reset to initial value
    assert photons['a'][1] == 2
    # Default initial value is nan
    assert np.isnan(photons['b'][1])

    # Add two columns at once, but in different format
    t.add_output_cols(photons, [{'name': '123', 'dtype': int}, 'text'])
    assert np.isnan(photons['text'][1])
    assert np.issubdtype(photons['123'].dtype, np.integer)


def test_add_idnum_col():
    '''Text another mechanism to add output columns'''
    class T(SSE):
        id_col = 'my_id'

    photons = Table({'a': [1, 2]})
    t = T()
    t.add_output_cols(photons)
    assert np.all(photons['my_id'] == -1)
    assert np.issubdtype(photons['my_id'].dtype, np.integer)


def test_output_col_none():
    '''Setting colnames to None, means that colum is not added.'''
    class T(SSE):
        id_col = None

    photons = Table({'a': [1, 2]})
    t = T()
    t.add_output_cols(photons)
    assert photons.colnames == ['a']


def test_tag():
    'check that tagging works'
    photons = Table({'a': [1, 2]})

    tag = TagVersion(qwer='texthere')
    photons = tag(photons, tag='check')
    assert photons.meta['qwer'] == 'texthere'
    assert photons.meta['tag'] == 'check'


def test_check_meta_consistent():
    check_meta_consistent({'ORIGIN': 'me', 'NOT_USED': 5},
                          {'ORIGIN': 'me'})
    with pytest.raises(KeyError,
                       match='CREATOR not found in both dicts.'):
        check_meta_consistent({'ORIGIN': 'me'}, {'ORIGIN': 'me'},
                              allow_missing=False)
    with pytest.raises(AssertionError):
        check_meta_consistent({'CREATOR': 'me'}, {'CREATOR': 'you'})


def test_check_energy_consistent():
    tab = Table({'energy': [5.1, 5.1, 5.1]})
    check_energy_consistent(tab)

    tab = Table({'energy': np.array([1., 2., 3.])})
    with pytest.raises(AssertionError):
        check_energy_consistent(tab)


# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest

from marxs.analysis.coords import ProjectOntoPlane, facet_table
from marxs.optics.base import OpticalElement
from marxs.math.utils import xyz2zxy
from marxs.simulator import Parallel
from marxs.analysis import ProjectOntoPlane


def test_facet_table():
    '''Make a facet table. At the same time, this tests ProjectOntoPlane'''
    gas = Parallel(elem_class=OpticalElement, elem_pos={'position':[[1, 1, 1], [2, 0, 4]]})
    ftab = facet_table(gas)
    proj2 = ProjectOntoPlane(orientation=xyz2zxy[:3, :3])
    ftab2 = facet_table(gas, proj2)
    assert ftab.colnames == ['facet_projx','facet_projy', 'facet_ang', 'facet_rad',
                             'facet', 'facet_x', 'facet_y', 'facet_z']
    assert len(ftab) == 2
    assert np.all(ftab['facet_projy'] == [1, 4])
    assert np.all(ftab2['facet_projy'] == [1, 0])
    assert np.all(ftab['facet_ang'] == pytest.approx([np.pi / 4, np.pi / 2]))
    assert np.all(ftab2['facet_ang'] == pytest.approx([np.pi / 4, 0]))
    assert dict(ftab2[1]) == {'facet_projx': 2.0,
                              'facet_projy': 0.0,
                              'facet_ang': 0.0,
                              'facet_rad': 2.0,
                              'facet': 1,
                              'facet_x': 2.0,
                              'facet_y': 0.0,
                              'facet_z': 4.0}



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import astropy.units as u
from astropy.table import Table
import pytest

from ..import ogip

def generate_RMF():
    '''Make an RMF for testing. Using one possible method
    to generate all the required inputs.'''
    width = Table({'wave': [10, 13, 14, 15, 16] * u.Angstrom,
              'sigma_wave':  [0.1, 0.2, 0.2, .2, .2] * u.Angstrom})
    rmf = ogip.RMF.from_Gauss_sigma(np.arange(10, 15, 0.1) * u.Angstrom,
                                    width)
    return rmf


def test_get_row_exact_energy():
    '''Make an RMF and check a row'''
    width = Table({'wave': [10, 13, 14, 15, 16] * u.Angstrom,
              'sigma_wave':  [0.1, 0.2, 0.2, .2, .2] * u.Angstrom})
    rmf = ogip.RMF.from_Gauss_sigma(np.arange(10, 15.5, 1) * u.Angstrom,
                                    width)
    energy = (10.5 * u.Angstrom).to(u.keV, equivalencies=u.spectral())
    row = rmf.row(energy)
    assert row['ENERG_LO'] * u.keV <= energy
    assert row['ENERG_HI'] * u.keV > energy
    assert row['N_GRP'] == 1
    assert np.min(row['MATRIX']) > 1e-6 # default threshold

    assert rmf.ebounds.meta['INSTRUME'] == 'unknown'


def test_threshold():
    width = Table({'wave': [10, 13, 14, 15, 16] * u.Angstrom,
              'sigma_wave':  [0.1, 0.2, 0.2, .2, .2] * u.Angstrom})
    rmf = ogip.RMF.from_Gauss_sigma(np.arange(10, 15, 0.1) * u.Angstrom,
                                    width)
    energy = (11.5 * u.Angstrom).to(u.keV, equivalencies=u.spectral())
    row = rmf.row(energy)
    assert row['N_CHAN'] == [19]
    assert np.min(row['MATRIX']) > 1e-6 # default threshold
    assert np.sum(row['MATRIX']) == pytest.approx(1)

    # Cut at a higher threshold, so fewer elements included
    rmf2 = ogip.RMF.from_Gauss_sigma(np.arange(10, 15, 0.1) * u.Angstrom,
                                     width, threshold=1e-4)
    row2 = rmf2.row(energy)
    assert row2['N_CHAN'][0] < 19
    assert np.min(row2['MATRIX']) > 1e-4
    assert np.sum(row['MATRIX']) > np.sum(row2['MATRIX'])

def test_arr_to_rmf_matrix_row():
    rmf = generate_RMF()

    arr = np.array([0, 1, 2, 0, 3.])
    ngrp, fchan, nchan, matrix = rmf.arr_to_rmf_matrix_row(arr, TLMIN_F_CHAN=0)
    assert ngrp == 2
    assert fchan == [1, 4]
    assert nchan == [2, 1]
    assert np.allclose(matrix, np.array([1, 2, 3]))

    ngrp, fchan, nchan, matrix = rmf.arr_to_rmf_matrix_row(arr, TLMIN_F_CHAN=1)
    assert ngrp == 2
    assert fchan == [2, 5]
    assert nchan == [2, 1]
    assert np.allclose(matrix, np.array([1, 2, 3]))




# Licensed under GPL version 3 - see LICENSE.rst
'''`X3D <https://www.web3d.org/x3d/what-x3d>`__ plotting backend

X3D is an open standard for interactive 3D models on the Web.
X3D models are stored in XML files and can be viewed in the
Jupyter notebook or on the web with different js libraries
without the installation of a plug-in.

This module implements the `~marxs.visualization.x3d.Scene`
class, which can represent itself as a complete HTML page to be
rendered directly or it can return the XML format to be saved
and used in other contexts.

Each plotting routine requires a `~marxs.visualization.x3d.Scene`
as input and objects are added to it (if no scene is passed in,
an empty one is generated.)

Under the hood, the XML is constructed using the
`x3d <https://pypi.org/project/x3d/>`__ Python package.
'''
from functools import wraps
from warnings import warn
import os
import urllib.request
import tempfile
from shutil import make_archive
import xml.etree.ElementTree as ET

import numpy as np
from astropy.utils.decorators import format_doc
from matplotlib.colors import Normalize
from matplotlib.pyplot import get_cmap
from x3d import x3d

from . import utils
from . import conf
from marxs.math import utils as mutils

__all__ = ['Scene',
           'empty_scene',
           'indexed_triangle_set',
           'surface',
           'triangulation',
           'box',
           'container',
           'plot_object',
           'plot_rays',
           'plot_registry',
           ]


doc_plot = '''
    {__doc__}

    Parameters
    ----------
    obj : `marxs.base.MarxsElement`
        The element that should be plotted.
    display : dict of None
        Dictionary with display settings.
    scene : `marxs.visualization.x3d.Scene` object
        A scene that this object is added to.
        If `None`, a new scene will be created.

    Returns
    -------
    scene : `marxs.visualization.x3d.Scene` object
        Scene with object added.
'''

class Scene(x3d.Scene):
    '''X3D Scene with added _repr_html_ for notebook output'''
    js_source = 'https://www.x3dom.org/download/x3dom.js'
    css_source = 'https://www.x3dom.org/download/x3dom.css'
    dimension_px = (600, 400)

    # see https://doc.x3dom.org/tutorials/animationInteraction/viewpoint/index.html
    # for how to add buttons for viewpoints

    def _repr_html_(self):
        root = ET.fromstring(self.XML())
        html = f"""
<html>
  <head>
    <meta http-equiv="X-UA-Compatible" content="IE=edge"/>
     <script type='text/javascript' src='{self.js_source}'> </script>
     <link rel='stylesheet' type='text/css' href='{self.css_source}'></link>
  </head>
  <body>
    <x3d width='{self.dimension_px[0]}px' height='{self.dimension_px[1]}px'>
      {self.XML()}
    </x3d>
    """
        for vp in root.findall('Viewpoint'):
            description = vp.get('description')
            html += f'<button onclick="document.querySelector(\'Viewpoint[description=\\\'{description}\\\']\').setAttribute(\'set_bind\',\'true\');">{description}</button>\n'
        html +="""
  </body>
</html>
"""
        return html

    def write_html_archive(self, base_name: str, format: str, *args, **kwargs) -> None:
        """Write to an HTML archive with local copies of js and css requirements.

        This zip file can be shared with others and opened in a web browser.
        In particular, it can be used for journal submission for journals
        that accept interactive js-based content such as ApJ or AJ.

        Parameters
        ----------
        base_name : str
            Name of the file to create, including the path,
            minus any format-specific extension.

        format : {"zip", "tar", "gztar", "bztar", "xztar"}
            The archive format, see `shutil.make_archive` for details.

        *args : tuple
            Other arguments are passed to `shutil.make_archive`.

        **kwargs : dict, optional
            Other keyword arguments are passed to `shutil.make_archive`
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            for f in (self.js_source, self.css_source):
                urllib.request.urlretrieve(
                    f, os.path.join(tmpdirname, os.path.basename(f))
                )
            try:
                old_js = self.js_source
                old_css = self.css_source
                self.js_source = os.path.basename(self.js_source)
                self.css_source = os.path.basename(self.css_source)
                with open(os.path.join(tmpdirname, "figure.html"), "w") as f:
                    f.write(self._repr_html_())
            finally:
                self.js_source = old_js
                self.css_source = old_css
            make_archive(base_name, format, root_dir=tmpdirname, *args, **kwargs)


def empty_scene(func):
    @wraps(func)
    def with_scene(*args, **kwargs):
        if 'scene' not in kwargs or kwargs['scene'] is None:
            kwargs['scene'] = Scene(children=[])
        return func(*args, **kwargs)
    return with_scene


def _diffuse_material(display):
    return x3d.Material(diffuseColor=display['color'],
                        transparency=1 - display.get('opacity', 1.))


def _format_points(xyz):
    return [tuple(np.round(p, conf.xyz_precision)) for p in xyz]


@empty_scene
def indexed_triangle_set(xyz, index, display, *, scene):
    '''Plot a set of triangles.

    Parameters
    ----------
    xyz : np.array
        Euclidean coordinates of nodes
    index : np.array
        index of nodes for each triangle
    display : dict of None
        Dictionary with display settings.
    scene : `marxs.visualization.x3d.Scene` object
        A scene that this object is added to.
        If `None`, a new scene will be created.

    Returns
    -------
    scene : `marxs.visualization.x3d.Scene` object
        Scene with object added.
    '''
    scene.children.append(x3d.Shape(appearance=x3d.Appearance(material=_diffuse_material(display)),
                     geometry=x3d.IndexedTriangleSet(coord=x3d.Coordinate(point=_format_points(xyz)),
                                                     index=[int(i) for i in index.reshape(-1, 3).flatten()],
                                                     solid=False, colorPerVertex=False)))


@empty_scene
@format_doc(doc_plot)
def surface(obj, display, *, scene):
    '''Plot a parametric surface.

    The parameter boundaries are taken from the ``coo1`` and ``coo2`` in the
    display dictionary. The plotting routine is generic. It calls the
    ``parametric_surface()`` method of the object that is plotted; see there
    for a detailed description of parameters.
    '''
    xyz = obj.geometry.parametric_surface(display.get('coo1', None),
                                          display.get('coo2', None),
                                          display)
    xyz = mutils.h2e(xyz)
    # number of faces. "-1" because last row just closes last face, but does not start a new one.
    # Each face has 4 vertices
    # [0, 2, 3, 1] is simply the right order to go around the polygon
    ind = np.arange(xyz.shape[0] * xyz.shape[1], dtype=int).reshape(xyz.shape[0], xyz.shape[1])
    coordIndex = np.stack([ind[:-1, :-1].flatten(),
                           ind[:-1, 1:].flatten(),
                           ind[1:, 1:].flatten(),
                           ind[1:, :-1].flatten(),
                            - np.ones((ind.shape[0] - 1) * (ind.shape[1] - 1),
                                      dtype=int)]).T
    coordIndex = coordIndex.flatten()
    scene.children.append(x3d.Shape(appearance=x3d.Appearance(material=_diffuse_material(display)),
                                    geometry=x3d.IndexedFaceSet(coord=
                                        x3d.Coordinate(point=_format_points(xyz.reshape((-1, 3)))),
                                        coordIndex=list(coordIndex),
                                        solid=False, colorPerVertex=False)))


@empty_scene
@format_doc(doc_plot)
def triangulation(obj, display, *, scene):
    '''Plot a plane, e.g. an aperture with an inner hole.'''
    xyz, index = obj.geometry.triangulate(display)
    indexed_triangle_set(xyz, index, display, scene=scene)
    return scene


@empty_scene
@format_doc(doc_plot)
def box(obj, display, *, scene):
    '''Plot a rectangular box for an object.

    By default, the box extends in x,y, and z direction. The display keyword
    "box-half" can be used to show only one half of the box, e.g. "+x" would
    show  the full extend in y and z direction, but only the lower half in
    x direction (such that rays coming from the +x direction are
    visible up to the interaction point).
    Use this for elements such as mirrors or detectors where
    photon interaction happens on the surface, not in the substrate.
    '''
    corners = utils.halfbox_corners(obj, display)
    shape = x3d.Shape(
        appearance=x3d.Appearance(material=_diffuse_material(display)),
        geometry=x3d.IndexedFaceSet(
            coord=x3d.Coordinate(point=_format_points(corners)),
            coordIndex=[
                0,
                2,
                3,
                1,
                -1,
                4,
                6,
                7,
                5,
                -1,
                0,
                4,
                6,
                2,
                -1,
                1,
                5,
                7,
                3,
                -1,
                0,
                4,
                5,
                1,
                -1,
                2,
                6,
                7,
                3,
                -1,
            ],
            solid=False,
            colorPerVertex=False,
        ),
    )
    scene.children.append(shape)

@empty_scene
@format_doc(doc_plot)
def container(obj, display=None, *, scene=None):
    '''Recursively plot objects contained in a container.'''
    for e in obj.elements:
        plot_object(e, display=None, scene=scene)
    return scene


@empty_scene
def plot_object(obj, display=None, *, scene=None, **kwargs):
    '''Plot any marxs object with using X3D as a backend.

    This method will inspect the object that is passed in and select the
    correct plotting method for its shape. The object is added to the
    x3d scene specified in the ``scene``.


    Parameters
    ----------
    obj : `marxs.base.MarxsElement`
        The element that should be plotted.
    display : dict of None
        Dictionary with display settings. If this is ``None``, ``obj.display``
        is used. If that is also ``None`` then the objects is skipped.
    scene : `marxs.visualization.x3d.Scene` object
        A scene that this object is added to.
        If `None`, a new scene will be created.
    kwargs
        All other parameters will be passed on to the individual plotting
        method.

    Returns
    -------
    scene : `marxs.visualization.x3d.Scene` object
        Scene with object added.
    '''
    utils.plot_object_general(plot_registry, obj, display, scene=scene, **kwargs)
    return scene


@empty_scene
def plot_rays(
    data, scalar=None, *, scene=None, cmap=get_cmap("viridis"), normalize_kwargs={}
):
    """Plot lines for simulated rays.

    Parameters
    ----------
    data : np.array of shape(n, N, 3)
        where n is the number of rays, N the number of positions per ray and
        the last dimension is the (x,y,z) of an Euclidean position vector.
    scalar : None or nd.array of shape (n,) or (n, N)
    scene : `marxs.visualization.x3d.Scene` object
        A scene that rays are added to.
        If `None`, a new scene will be created.
    normalize_kwargs : dict
        Keyword arguments for the normalization of the scalar quantity
        see `matplotlib.colors.Normalize` for accepted keywords.

    Returns
    -------
    scene : `marxs.visualization.x3d.Scene` object
        Scene with object added.
    """
    # The number of points per line
    N = data.shape[1]
    # number of lines
    n = data.shape[0]

    if scalar is None:
        # Color all rays the same
        scalar = np.ones(n)
    elif scalar.shape == (n, ):
        pass
    elif scalar.shape == (n, N):
        warn('Color Per Vertex does not yet work in the X3D library used. Using only color for first node in each ray.')
        scalar = scalar[0, :]
    else:
        raise ValueError('Scalar quantity for each point must have shape ({0},) or ({0}, {1})'.format(n, N))

    scalar = Normalize(**normalize_kwargs)(scalar)
    scalarset = set(scalar)
    # If scalar is a float quantity it makes smaller X3D file to bin and have e.g. no more than 100 colors.
    # Not hard to implement, but punt for now since I don't know if that optimization is needed in
    # practice.
    for s in scalarset:
        color = cmap(s)
        color = tuple(np.round(color, conf.color_precision))
        # color is RGBA, but I have not figured out the alpha in X3D, so just drop that
        # and use RGB
        # Also, round to two digits to reduce X3D filesize
        # color = x3d.Color(tuple(np.round(color[:3], 2)))
        ind = scalar == s

        lines = x3d.Shape(
            appearance=x3d.Appearance(material=x3d.Material(emissiveColor=color[:3])),
            geometry=x3d.LineSet(vertexCount=[N] * ind.sum(),
                                 # Rounding to two post-comma digits (i.e. 0.1 mm in MARXS default units)
                                 # to keep file size down
                                coord=x3d.Coordinate(point=_format_points(data[ind, :].reshape(-1, 3))),
                                )
        )
        scene.children.append(lines)
    return scene



plot_registry = {'triangulation': triangulation,
                 'box': box,
                 'container': container,
                 'surface': surface,
                 }


# Licensed under GPL version 3 - see LICENSE.rst
import math
import warnings

import astropy.units as u
import numpy as np
from transforms3d.affines import decompose44
from scipy.stats import norm

from .base import FlatOpticalElement, OpticalElement
from ..utils import SimulationSetupWarning
from ..math.geometry import Cylinder


__all__ = ['FlatDetector', 'CircularDetector', 'CCDRedistNormal']


class FlatDetector(FlatOpticalElement):
    '''Flat detector with square pixels

    Processing the a photon with this detector adds four columns to the photon
    properties:

    - ``det_x``, ``det_y``: Event position on the detector in mm
    - ``detpix_x``, ``detpix_y``: Event position in detector coordinates, where
      (0, 0) is on the corner of the chip.

    The (x,y) coordinate system on the chip is such that it falls on the (y,z)
    plane of the global x,y,z coordinate system (it would be more logical to
    call the chip system (y,z), but traditionally that is not how chip
    coordinates are named). The pixel in the corner has coordinates (0, 0) in
    the pixel center.

    Parameters
    ----------
    pixsize : float
        size of pixels in mm

    ignore_pixel_warning : bool
        ignore warning if derived pixel number is not close to an integer

    kwargs :
       see `args for optical elements`

    '''

    loc_coos_name = ['det_x', 'det_y']
    '''name for output columns that contain the interaction point in local coordinates.'''

    detpix_name = ['detpix_x', 'detpix_y']
    '''name for output columns that contain this pixel number.'''



    display = {'color': (1.0, 1.0, 0.),
               'shape': 'box',
               'box-half': '+x'}

    def __init__(self, pixsize=1, ignore_pixel_warning=False, **kwargs):
        self.pixsize = pixsize
        super().__init__(**kwargs)
        t, r, zoom, s = decompose44(self.pos4d)
        self.npix = [0, 0]
        self.centerpix = [0, 0]
        for i in (0, 1):
            z  = zoom[i + 1]
            self.npix[i] = int(np.round(2. * z / self.pixsize))
            if (np.abs(2. * z / self.pixsize - self.npix[i]) > 1e-2) and not ignore_pixel_warning:
                warnings.warn('Detector size is not an integer multiple of pixel size in direction {0}. It will be rounded.'.format('xy'[i]), SimulationSetupWarning)
            self.centerpix[i] = (self.npix[i] - 1) / 2

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        detx = intercoos[intersect, 0] / self.pixsize + self.centerpix[0]
        dety = intercoos[intersect, 1] / self.pixsize + self.centerpix[1]
        return {self.detpix_name[0]: detx, self.detpix_name[1]: dety}


class CircularDetector(OpticalElement):
    '''A detector shaped like a ring or tube.

    This detector is shaped like a tube. The form is a circle in the xy plane
    and and flat along the z-direction.  While most CCDs are flat in practice,
    the `CircularDetector` simulates a setup that can follow the Rowland circle
    geometry exactly which is useful, e.g. to study the resolution of a
    spectrograph without worrying about the details of the detector geometry.

    Parameters
    ----------
    position, orientation, zoom, pos4d : see description of `pos4d`
        The radius of the tube is given by the ``zoom`` keyword, see
        `pos4d`.  Use ``zoom[0] == zoom[1]`` to make a circular
        tube. ``zoom[0] != zoom[1]`` gives an elliptical
        profile. ``zoom[2]`` sets the extension in the z direction.
    pixsize : float size of pixels in mm

    '''
    loc_coos_name = ['det_phi', 'det_y']

    detpix_name = ['detpix_x', 'detpix_y']
    '''name for output columns that contain this pixel number.'''

    display = {'color': (1.0, 1.0, 0.),
               'opacity': 0.7,
               }

    centerpix = [0, 0]

    default_geometry = Cylinder

    def __init__(self, pixsize=1, **kwargs):
        self.pixsize = pixsize
        super().__init__(**kwargs)

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        detx = intercoos[intersect, 0] * self.geometry['R'] / self.pixsize + self.centerpix[0]
        dety = intercoos[intersect, 1] / self.pixsize + self.centerpix[1]
        return {self.detpix_name[0]: detx, self.detpix_name[1]: dety}



class CCDRedistNormal(FlatOpticalElement):
    """Redistribute energies according to a normal distribution

    No detector has infinite energy resolution. This class redistributes the
    the energy according to a normal distribution.

    This class serves a dual role. It can be used as an optical element within an
    instrument to cheaply simulate a CCD response (not including effects like frame
    transfer or CTI), but it can also be used in the OSIP generators in
    `marxs.reduction.osip`.

    Parameters
    ----------
    tab_width : astropy.table.QTable
        Table with columns "energy" and either "sigma" or "FWHM"

    Notes
    -----
    EXPERIMENTAL.
    For now, I'm just putting in methods one by one as I need them to help
    me develop how the stable API should look like.
    In principle this class could inherit from `scipy.stat.norm`
    or maybe from a sherpa norm1d distribution or from astropy.modelling
    (which should be quantity aware already).
    Eventually, this might be a good application of a metaclass
    e.g. as in https://stackoverflow.com/questions/11349183/how-to-wrap-every-method-of-a-class
    to handles scale and loc arguments automatically in the way shown
    below, but for now it's easier ot just copy and paste that wrapping
    a few times as needed.


    Also, this is for instance methods. I think "norm" might use class methods
    so that's just one step more complicated...

    """

    fwhm2sig = 2 * math.sqrt(2 * math.log(2))

    energycol = "energy_detected"

    def __init__(self, **kwargs):
        self.tab_width = kwargs.pop("tab_width")
        if 'sigma' not in self.tab_width.colnames:
            self.tab_width['sigma'] = self.tab_width['FWHM'] / self.fwhm2sig
        super().__init__(**kwargs)

    @u.quantity_input(energy=u.keV, equivalencies=u.spectral())
    def sig_ccd(self, energy):
        '''Return the Gaussian sigma of the width of the CCD resolution

        Parameters
        ----------
        energy : `~astropy.units.quantity.Quantity`
            True photon energy.

        Returns
        -------
        sigma : `~astropy.units.quantity.Quantity`
            Width of the Gaussian
        '''
        return np.interp(energy.to(u.keV, equivalencies=u.spectral()),
                         self.tab_width['energy'],
                         self.tab_width['sigma'])

    @u.quantity_input(x=u.keV, loc=u.keV, equivalencies=u.spectral())
    def cdf(self, x, loc):
        scale = self.sig_ccd(loc)
        return norm().cdf(((x - loc) / scale).decompose())

    @u.quantity_input(loc=u.keV, equivalencies=u.spectral())
    def interval(self, alpha, loc):
        scale = self.sig_ccd(loc)
        return np.broadcast_to(norm.interval(alpha), (len(loc), 2)).T * scale + loc

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        # Implicit unit of photon energy is keV, but here we need it written out.
        phot_en = u.Quantity(photons["energy"])
        det_en = norm.rvs(
            loc=phot_en,
            scale=self.sig_ccd(phot_en).to(u.keV, equivalencies=u.spectral()),
        )
        return {self.energycol: det_en}



import copy

import astropy.units as u
import numpy as np
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename

from marxs import analysis, design, optics, simulator
from marxs.base import MarxsElement, TagVersion
from marxs.design import tolerancing as tol
from marxs.design.bendgratings import NumericalChirpFinder, chirp_gratings
from marxs.math.geometry import Cylinder
from marxs.missions.arcus.arcus import FiltersAndQE
from marxs.missions.lynx.lynx import LynxGAS, add_rowland_to_conf
from marxs.missions.lynx.mirror import MetaShellAperture, PerfectLensSegment
from marxs.missions.mitsnl.catgrating import (CATL1L2Stack,
                                              InterpolateEfficiencyTable,
                                              NonParallelCATGrating,
                                              catsupportbars)
from marxs.optics import FlatDetector, OrderSelector
from marxs.simulator import Propagator

tagversion = TagVersion(SATELLIT='AXIS')

# CAT gratings tabluated data
order_selector_Si = InterpolateEfficiencyTable(Table.read(
    get_pkg_data_filename('data/Si_efficiency_5_7.dat',
                          package='marxs.missions.mitsnl'),
    format='ascii.ecsv'))
order_selector_Si.coating = 'None'
order_selector_Pt = InterpolateEfficiencyTable(Table.read(
    get_pkg_data_filename('data/SiPt_efficiency_5_7.dat',
                          package='marxs.missions.mitsnl'),
    format='ascii.ecsv'))
order_selector_Pt.coating = 'Pt'


l1_order_selector = OrderSelector(orderlist=np.array([-4, -3, -2, -1, 0, 1, 2, 3, 4]),
                                  p=np.array([0.006, 0.0135, 0.022, 0.028, 0.861, 0.028, 0.022, 0.0135, 0.006]))
'''Simple order selector for diffraction on L1.

The numbers in the array are calculated for the 2018 Arcus gratings
assuming the L1 structure is independent from the grating membrane itself
(which is not true, but a valid first approximation.)
For AXIS, we are using 50% deeper gratings, yet the L1 bars are thinner.
Need better numbers for that.
'''


conf = {'metashellgeometry': Table.read(get_pkg_data_filename('AXIS_metashellgeom.dat'),
                                        format='ascii.ecsv'),
        # This gives a PSF HPD around 1 arcsec.
        # The nominal 1 arcsec includes alignment, but until I code that up, I just makes 
        # PSF a little wider here.
        # 4e-6 and 2e-6 work well otherwise.
        'inplanescatter': 5e-6 * u.rad,
        'perpplanescatter': 2.5e-6 * u.rad,
        'mirrorcoating': 'ir_reflectivity.dat',
        'mirrorgeometricscaling': 0.61,
        'blazeang': np.deg2rad(1.6),
        'focallength': 9000.,
        'design_tilted_torus': 8.6e3, 
        'alphafac': 2.2,
        'betafac': 4.4,
        'grating_size': np.array([30., 60.]),
        'grating_frame': 2.,
        'det_kwargs': {'y_range': [400, 600],
                       'd_element': [24.576 * 2 + 0.824 * 2 + 0.5,
                                     24.576 + 0.824 * 2 + 0.5],
                       'elem_class': optics.FlatDetector,
                       'elem_args': {'zoom': [1, 24.576, 12.288],
                                     'pixsize': 0.024,
                                     # orientation flips around CCDs so that det_x increases
                                     # with increasing x coordinate
                                    'orientation': np.array([[-1, 0, 0],
                                                             [0, -1, 0],
                                                             [0, 0, +1]])
                                     }},
        'gas_kwargs': {'parallel_spec': np.array([0., 1., 0., 0.]),
                       'normal_spec': np.array([0, 0, 0, 1]),
                       'guess_distance': 9e3,
                       'elem_class': CATL1L2Stack,
                       'elem_args': {'d': 2e-4,
                                     'order_selector': order_selector_Si,
                                     'l1_dims': {'bardepth': 5.7 * u.micrometer,
                                                 'period': 0.005 * u.mm,
                                                 'barwidth': 0.0007 * u.mm},
                                     'l2_dims': {'bardepth': 0.5 * u.mm,
                                                 'period': 0.966 * u.mm,
                                                 'barwidth': 0.07 * u.mm},
                                     # Set sigma to 0. Effectively disables this
                                     # factor.
                                     'qualityfactor': {'d': 200. * u.um,
                                                       'sigma': 1.5 * u.um},
                                     'l1_order_selector': l1_order_selector,
                                     },
                       },
    }


add_rowland_to_conf(conf)

conf_6060 = copy.copy(conf)
conf_6060['grating_size'] = np.array([60., 60.])

conf_chirp = copy.copy(conf)
conf_chirp['grating_size'] = np.array([80., 160.])
conf_chirp['chirp_energy'] = 0.6
# Use of a non-integer order makes no sense physically
# but is just a numerical tool to optimize at the blaze peak
conf_chirp['chirp_order'] = -5.4



class RowlandDetArray(design.rowland.RectangularGrid):
    def __init__(self, conf):
        y_offset = conf['rowland'].geometry['center'][1]
        super().__init__(rowland=conf['rowland'],
                         y_range=np.array(conf['det_kwargs']['y_range']) - y_offset,
                         d_element=conf['det_kwargs']['d_element'],
                         elem_args=conf['det_kwargs']['elem_args'],
                         elem_class=conf['det_kwargs']['elem_class'],
                         guess_distance=25.,
                         id_col='CCD_ID')


imaging = FlatDetector(zoom=[1, 15, 15], pixsize=0.008)
# zero-order detector based on https://axis.astro.umd.edu/images/proposal.pdf
imaging.loc_coos_name = ['imaging_x', 'imaging_y']
imaging.detpix_name = ['imaging_x', 'imaging_y']


# Place an additional detector on the Rowland circle.
detcirc = optics.CircularDetector(geometry=Cylinder.from_rowland(conf['rowland'],
                                                                 width=50,
                                                                 rotation=np.pi,
                                                                 kwargs={'phi_lim':[-np.pi/2, np.pi/2]}))
detcirc.loc_coos_name = ['detcirc_phi', 'detcirc_y']
detcirc.detpix_name = ['detcircpix_x', 'detcircpix_y']

flatdet = FlatDetector(zoom=[1, 1e5, 1e5])
flatdet.loc_coos_name = ['detinf_x', 'detinf_y']
flatdet.detpix_name = ['detinfpix_x', 'detinfpix_y']
flatdet.display['shape'] = 'None'


class MirrorGeometricScaling(MarxsElement):
    def __init__(self, conf):
        self.scale = conf['mirrorgeometricscaling']

    def __call__(self, photons):
        photons['probability'] = self.scale * photons['probability']
        return photons


class PerfectAXIS(simulator.Sequence):
    '''Default Definition of Lynx without any misalignments'''
    def add_mirror(self, conf):
        return [MetaShellAperture(conf), 
                PerfectLensSegment(conf, position=[conf['focallength'], 0, 0],
                zoom=[1, 1000, 1000]),
                MirrorGeometricScaling(conf),
                optics.RadialMirrorScatter(inplanescatter= conf['inplanescatter'],
                                           perpplanescatter=conf['perpplanescatter'],
                                           position=[conf['focallength'], 0, 0],
                                           zoom=[1, 1000, 1000])]

    def add_gas(self, conf):
        return [LynxGAS(conf),
                catsupportbars]

    def add_detectors(self, conf):
        '''Add detectors to the element list

        This is a separate function that is called from __init__ because all
        detectors need different parameters. Placing this specific code in its own
        function makes it easy to override for derived classes.
        '''
        proj2 = analysis.ProjectOntoPlane()
        proj2.loc_coos_name = ['projcirc_y', 'projcirc_z']
        return [RowlandDetArray(conf),
                imaging,
                analysis.ProjectOntoPlane(),
                Propagator(distance=-1000),
                detcirc,
                proj2,
                Propagator(distance=-1000),
                flatdet]

    def post_process(self):
        self.KeepPos = simulator.KeepCol('pos')
        return [self.KeepPos]

    def __init__(self, conf=conf, gratings=True, **kwargs):
        elem = self.add_mirror(conf)
        if gratings:
            elem.extend(self.add_gas(conf))
        elem.append(FiltersAndQE())
        elem.extend(self.add_detectors(conf))
        elem.append(tagversion)

        super().__init__(elements=elem,
                         postprocess_steps=self.post_process(),
                         **kwargs)
        if ('chirp_energy' in conf) and ('chirp_order' in conf):
            gratings = self.elements_of_class(NonParallelCATGrating, subclass_ok=False)
            opt = NumericalChirpFinder(detcirc,
                                       order=conf['chirp_order'],
                                       energy=conf['chirp_energy'],
                                       d=conf['gas_kwargs']['elem_args']['d'])
            chirp_gratings(gratings, opt, conf['gas_kwargs']['elem_args']['d'])


class AXISForPlot(PerfectAXIS):
    def add_detectors(self, conf):
        zero = FlatDetector(zoom=[1, 15, 15])
        zero.loc_coos_name = ['zero_x', 'zero_y']
        zero.detpix_name = ['zero_x', 'zero_y']
        return [RowlandDetArray(conf),
                imaging]


class AXIS(PerfectAXIS):
    def __init__(self, conf=conf, **kwargs):
        super().__init__(conf=conf, **kwargs)
        for row in conf['alignmentbudget']:
            elem = self.elements_of_class(row[0])
            if row[1] == 'global':
                tol.moveglobal(elem, *row[3])
            elif row[1] == 'individual':
                tol.wiggle(elem, *row[3])
            else:
                raise NotImplementedError('Alignment error {} not implemented'.format(row[1]))




# Licensed under GPL version 3 - see LICENSE.rst
from os.path import join as pjoin
import logging
import os
from abc import ABC
from glob import glob

import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u

from . import ogip
from marxs.analysis.plots import OrderColor


logger = logging.getLogger(__name__)

__all__ = ['OSIPBase',
           'FixedWidthOSIP', 'FixedFractionOSIP', 'FractionalDistanceOSIP',
           ]


class OSIPBase(ABC):
    """Modify ARF files to account for order-sorting effects

    This is a base class that implements order sorting and integration of
    the probabilities (OSIP). This includes order-sorting of a photon list
    (not yet implemented) and  methods to modify ARF files
    to account for order-sorting. Different diffraction orders fall on
    the same physical space on the CCD. The CCD energy resolution can
    then be used to assign photons to specific grating
    orders. However, since the CCD energy resolution is finite, this
    assignment is not perfect. Some photons may fall outside of the
    nominal energy used for order sorting and thus be lost, other may
    fall onto the energy assigned to the different order. This class
    modifies ARF files to correct for the integrated probability that a photon
    falls inside the order-sorting region.

    Derived classes should override either `osip_tab` or `osip_range`. The
    other of the two can always be calculated from whatever one is given.

    Parameters
    ----------
    offset_orders : list of int
        Offset from main order that is relevant.

    ccd_redist : Redistribution object
        Function that return the width the Gaussian sigma of the CCD
        resolution, given an input energy.
        TODO: Better docs here
    """
    osip_description = 'Not implemented'
    '''String to be added to FITS header in OSIP keyword for ARFs written.'''

    def __init__(self, ccd_redist, offset_orders=[-1, 0, 1],
                 filename_from_meta=ogip.filename_from_meta):
        self.offset_orders = offset_orders
        self.ccd_redist = ccd_redist
        self.filename_from_meta = filename_from_meta

    @u.quantity_input(chan_mid=u.keV, equivalencies=u.spectral())
    def osip_tab(self, chan_mid, order) -> u.eV:
        '''Calculate the boundaries of an order-sorting region

        Parameters
        ----------
        energy : `~astropy.units.quantity.Quantity`
            Energy (or wavelength actually) for which the OSIP should
            be calculated, e.g. the mid-point of a channel
        order : int
            Diffraction order.

        Returns
        -------
        osip_tab : `~astropy.units.quantity.Quantity`
            The shape is an array that contains, for each point in `energy`,
            the width of the OSIP region, as measured from the nominal
            energy. If, for example, the energy is 1 keV and ``osip_tab``
            is `[.1, .2] * u.keV`, then the osip is [.9, 1.2] keV.
        '''
        osiprange = self.osip_range(chan_mid, order)
        osiprange = osiprange.to(u.eV, equivalencies=u.spectral())
        E = chan_mid.to(u.eV, equivalencies=u.spectral())
        return E + osiprange * np.array([-1, 1])[:, np.newaxis]

    @u.quantity_input(chan_mid=u.keV, equivalencies=u.spectral())
    def osip_range(self, chan_mid, order) -> u.eV:
        '''Calculate the boundaries of an order-sorting region

        This method returns the boundaries of an order-sorting region in
        energy space.
        This is very similar to `osip_tab`, which gives the width of the
        region. Essentially, this method calls `osip_tab` and add the energy
        so that the result contains the lower and upper bound.

        Parameters
        ----------
        chan_mid : `~astropy.units.quantity.Quantity`
            Energy or wavelength for which the OSIP should
            be calculated, e.g. the mid-point of a channel
        order : int
            Diffraction order.

        Returns
        -------
        osip : `~astropy.units.quantity.Quantity`
            The shape is an array that contains, for each point in `energy`,
            the lower and upper bound of the OSIP region.
        '''
        osiptab = self.osip_tab(chan_mid, order)
        osiptab = osiptab.to(u.eV, equivalencies=u.spectral())
        E = chan_mid.to(u.eV, equivalencies=u.spectral())
        return E + osiptab * np.array([-1, 1])[:, np.newaxis]

    @u.quantity_input(chan_mid_nominal=u.keV, equivalencies=u.spectral())
    def osip_factor(self, chan_mid_nominal, o_nominal, o_true):
        '''Calculate the relative effective area after order sorting.

        This method can be used to calculate a fraction fo

        Parameters
        ----------
        chan_mid_nominal : `~astropy.units.quantity.Quantity`
            Energy or wavelength of the nominal order for which the OSIP should
            be calculated, e.g. the mid-point of a channel
        o_nominal : int
            Nominal diffraction order, i.e. the order that is to be extracted
            from the CCD.
        o_true : int
            True diffraction order.

        Returns
        -------
        osip : `~astropy.units.quantity.Quantity`
            The shape is an array that contains, for each point in `energy`,
            the lower and upper bound of the OSIP region.
        '''

        if np.sign(o_nominal) != np.sign(o_true):
            return 0.

        osiprange = self.osip_range(chan_mid_nominal, o_nominal)
        Etrue = chan_mid_nominal.to(u.eV, equivalencies=u.spectral()) * \
            (o_true / o_nominal)
        upper = self.ccd_redist.cdf(osiprange[1, :], loc=Etrue)
        lower = self.ccd_redist.cdf(osiprange[0, :], loc=Etrue)
        return upper - lower

    def apply_osip(self, inputarf, outpath, order, outroot='',
                   overwrite=False):
        '''Modify an ARF to account for incomplete order sorting

        This function reads an input ARF file, which contains the
        effective area for a grating order in Arcus. For a given order
        sorting window, it then calculates what fraction of the
        photons is lost. For example, if the order-sorting regions can be
        is chosen to contain  90% energy fraction, then the new ARF values
        will be 0.9 (the integrated probability over the order-sorting region)
        times the input ARF.

        If the ``order`` of the new ARF differs from the order of the
        input ARF, then the new ARF is representative of order
        confusion, e.g. is shows how many photons of order 4 are
        sorted into the order=5 grating spectrum.

        Parameters
        ----------
        inputarf : string
            Filename and path of input arf
        outpath : string
            Location where the output arfs are deposited
        order : int
            Nominal order. (The true order of the input arf is taken from the
            input arf header data.)
        outfile : string
            path and filename for output file
        overwrite : bool
            Overwrite existing files?
        '''
        arf = ogip.ARF.read(inputarf)
        try:
            arf['SPECRESP'] = arf['SPECRESP'] / arf['OSIPFAC']
            logger.info(f'{inputarf} already has OSIP applied, reverting ' +
                        'before applying new OSIP.')
        except KeyError:
            pass

        m = int(arf.meta['ORDER'])

        energies = 0.5 * (arf['ENERG_LO'] + arf['ENERG_HI'])
        osip_fac = self.osip_factor(energies / m * order, order, m)
        arf['SPECRESP'] = osip_fac * arf['SPECRESP']
        arf['OSIPFAC'] = osip_fac

        arf.meta['INSTRUME'] = f'ORDER_{order}'
        arf.meta['OSIP'] = self.osip_description
        arf.meta['TRUEORD'] = f'{m}'
        arf.meta['CCDORDER'] = f'{order}'
        # some times there is no overlap and all elements become 0
        if np.all(arf['SPECRESP'] == 0):
            logger.warning(f'True refl order {m} does not contribute to ' +
                        f'CCDORDER {order}. ' +
                        'Writing ARF with all entries equal to zero.')
        os.makedirs(outpath, exist_ok=True)
        arf.write(pjoin(outpath, outroot +
                        self.filename_from_meta('arf', **arf.meta)),
                  overwrite=overwrite)


    def plot_osip(self, ax, grid, order, **kwargs):
        '''Plot banana plot with OSIP region marked.

        Parameters
        ----------
        ax : `matplotlib.axes._subplots.AxesSubplot`
            The axes into which the banana is plotted.
        grid : `~astropy.units.quantity.Quantity`
            Wavelength grid in m lambda
        order : int
            Order number
        kwargs :
            Any other parameters are passed to ``plt.plot``
        '''
        grid = grid.to(u.Angstrom, equivalencies=u.spectral())
        en = (grid / np.abs(order)).to(u.keV, equivalencies=u.spectral())

        ohw = self.osip_tab(en, order)

        line = ax.plot(grid, en, label=order, **kwargs)
        ax.fill_between(grid.value,
                        (en - ohw[0, :]).value,
                        (en + ohw[1, :]).value,
                        color=line[0].get_color(), alpha=.2,
                        label='__no_legend__')
        ax.set_xlabel(f'$m\\lambda$ [{grid.unit.to_string("latex_inline")}]')
        ax.set_ylabel('CCD energy [keV]')
        ax.set_xlim([grid.value.min(), grid.value.max()])
        ax.legend()
        ax.set_title('Order sorting regions')

    def plot_mixture(self, ax, grid, order):
        '''Plot relative contribution of main order and interlopers.

        Parameters
        ----------
        ax : `matplotlib.axes._subplots.AxesSubplot`
            The axes into which the lines are plotted.
        grid : `~astropy.units.quantity.Quantity`
            Wavelength grid in m lambda
        order : int
            Order number
        '''
        grid = grid.to(u.Angstrom, equivalencies=u.spectral())
        ax.axhspan(1, 2, facecolor='r', alpha=.3, label='extractions overlap')
        en = (grid / np.abs(order)).to(u.keV, equivalencies=u.spectral())
        cm = self.osip_factor(en, order, order)
        ax.plot(grid, cm, label='main order', lw=2)
        for o in self.offset_orders:
            if o == 0:
                continue
            coffset = self.osip_factor(en, order, order + o)
            ax.plot(grid, coffset, label=f'contam {o}',
                    # Different linestyle to avoid overlapping lines in plot
                    ls={-1: '-', +1: ':'}[np.sign(o)]
                    )
            cm += coffset
        ax.plot(grid, cm, 'k', label='sum', lw=3)
        ax.set_xlabel(f'$m\\lambda$ [{grid.unit.to_string("latex_inline")}]')
        ax.set_ylabel('Fraction of photons in OSIP')
        ax.set_title(f'Order {order}')
        ax.set_xlim([grid.value.min(), grid.value.max()])
        ax.set_ylim(0, cm.max() * 1.05)
        ax.legend()

    def plot_summary(self, inputarf, orders, outpath, outroot=''):
        '''Write summary plot to directory with ARFs

        Parameters
        ----------
        inputarf : string
            Path to one input ARF. The energy grid for the plot
            is taken from that ARF.
        outpath : string
            Location where the output ARFs are deposited
        outroot : string
            prefix for output filename
        '''
        arf = ogip.ARF.read(inputarf)
        bin_mid = 0.5 * (arf['ENERG_LO'] + arf['ENERG_HI'])
        grid = bin_mid.to(u.Angstrom, equivalencies=u.spectral())
        fig, axes = plt.subplots(ncols=2, figsize=(8, 4))

        oc = OrderColor(max_order=np.max(np.abs(orders)))

        for order in orders:
            self.plot_osip(axes[0], grid * abs(arf.meta['ORDER']), order,
                           **oc(order))
        # pick the middle order for plotting purposes
        o_mid = orders[len(orders) // 2]
        self.plot_mixture(axes[1], grid * abs(arf.meta['ORDER']), o_mid)
        fig.subplots_adjust(wspace=.3)
        fig.savefig(pjoin(outpath, outroot + 'OSIP_regions.pdf'),
                    bbox_inches='tight')

    def plot_summed_arf(self, path):
        '''Generate a plot of the summed effective area.

        This plot is not meant to convey detailed information, but rather
        serve as a check for verification.

        Parameters
        ----------
        path : string
            Path to location of ARFs. The plot is written to the same file.
        '''
        wavegrid = np.arange(0, 60, .1) * u.Angstrom
        specresp = np.zeros(len(wavegrid)) * u.cm**2
        arflist = glob(pjoin(path, '*.arf'))
        for f in arflist:
            arf = ogip.ARF.read(f)
            # Interpolate on common wavelength grid. Good enough for plotting.
            # Interp requires sorted input
            # but ARFs are reverse-sorted in wavelength
            specresp += np.interp(wavegrid, arf['BIN_LO'][::-1],
                                  arf['SPECRESP'][::-1])

        fig, ax = plt.subplots()
        ax.plot(wavegrid, specresp)
        ax.set_xlabel(f'$\\lambda$ [{wavegrid.unit.to_string("latex_inline")}]')
        ax.set_ylabel(f'effective area [{specresp.unit.to_string("latex_inline")}]')
        ax.set_title('Combined effective area of all dispersed orders')
        fig.savefig(pjoin(path, 'ARF_sum.pdf'),
                    bbox_inches='tight')

    def apply_osip_all(self, inpath, outpath, orders,
                       inroot='', outroot='',
                       overwrite=False,
                       filename_from_meta_kwargs={}):
        '''Apply OSIP to many arfs at once

        This routine iterates over orders and offset orders to produce
        arfs that describe the contamination due to insufficient order
        sorting.  When a single order is extracted (e.g. order 5), but
        the CCD resolution is insufficient to really separate the
        orders well, then some photons from order 4 and 6 might end up
        in the extracted spectrum. This function generates the
        appropriate arfs.

        Input arfs need to follow the arcus filename convention and
        all be located in the same directory. In addition to the ARFs
        with order-sorting applied, this method also places an
        overview plot in the output directory.

        Parameters
        ----------
        inpath : string
            Directory with input ARFs.
        outpath : string
            Location where the output ARFs are deposited
        orders : list of int
            Nominal CCD orders to be processed
        inroot : string
            prefix for input filename
        outroot : string
            prefix for output filename
        overwrite : bool
            Overwrite existing files?
        filename_from_meta_kwargs : dict
            Any additional kwargs needed to run ``self.filename_from_meta``
            and thus pick the correct input arfs in the `inpath`
        '''
        goodarf = None
        for order in orders:
            for t in self.offset_orders:
                # No contamination by zeroth order or by orders on the other
                # side of the zeroth order
                if (order + t != 0) and (np.sign(order) == np.sign(order + t)):
                    inputarf = pjoin(inpath, inroot +
                                     self.filename_from_meta(filetype='arf',
                                                        ORDER=order + t,
                                                        **filename_from_meta_kwargs))
                    try:
                        self.apply_osip(inputarf, outpath, order,
                                        outroot=outroot, overwrite=overwrite)
                        goodarf = inputarf
                    except FileNotFoundError:
                        logger.info(f'Skipping order: {order}, offset: {t} ' +
                                    'because input arf not found')
                        continue

        # The second condition checks that at least one ARF was written
        if goodarf is not None:
            self.plot_summary(goodarf, orders, outpath, outroot)
            self.plot_summed_arf(outpath)


class FixedWidthOSIP(OSIPBase):
    """Modify ARF files to account for order-sorting effects

    This class implements order sorting and integration of
    the probabilities (OSIP) for order-sorting regions that have a fixed witdh
    in energy.
    This includes order-sorting of a photons list
    (not yet implmented) and  methods to modify ARF files
    to account for order-sorting. Different diffraction orders fall on
    the same physical space on the CCD. The CCD energy resolution can
    then be used to assign photons to specific grating
    orders. However, since the CCD energy resolution is finite, this
    assingment is not perfect. Some photons may fall outside of the
    nominal energy used for order sorting and thus be lost, other may
    fall onto the energy assgined tothe different order. This class
    modifies ARF files to correct for the integrated probability that a photon
    falls inside the order-sorting region.

    Parameters
    ----------
    halfwidth : `astropy.units.quantity.Quantity`
        Half-wdith of the order sorting region. The same width is used for all
        wavelength.
    offset_orders : list of int
        Offset from main order that is relevant.
    ccd_redist : callable
        Function that returns the width the Gaussian sigma of the CCD
        resolution, given an input energy.
    """
    def __init__(self, halfwidth, **kwargs):
        self.halfwidth = halfwidth
        super().__init__(**kwargs)

    @property
    def osip_description(self):
        return str(self.halfwidth)

    @u.quantity_input(chan_mid_nominal=u.keV, equivalencies=u.spectral())
    def osip_tab(self, chan_mid_nominal, order):
        return np.broadcast_to(self.halfwidth,
                               (2, len(chan_mid_nominal)), subok=True)


class FixedFractionOSIP(OSIPBase):
    '''Modify ARF files to account for order-sorting effects

    This class implements order sorting and integration of
    the probabilities (OSIP) for order sorting regions that contain a fixed
    fraction of the photons in energy space.
    This includes order-sorting of a photons list
    (not yet implmented) and  methods to modify ARF files
    to account for order-sorting. Different diffraction orders fall on
    the same physical space on the CCD. The CCD energy resolution can
    then be used to assign photons to specific grating
    orders. However, since the CCD energy resolution is finite, this
    assingment is not perfect. Some photons may fall outside of the
    nominal energy used for order sorting and thus be lost, other may
    fall onto the energy assgined tothe different order. This class
    modifies ARF files to correct for the integrated probability that a photon
    falls inside the order-sorting region.

    Parameters
    ----------
    fraction : float
        Number (between 0 and 1) that determins which fraction of the CCD
        energy distriution should be covered by the order sorting regions.
        The same width is used for all
        wavelength.
    offset_orders : list of int
        Offset from main order that is relevant.
    ccd_redist : callable
        Function that return the width the Gaussian sigma of the CCD
        resolution, given an input energy.
    '''
    def __init__(self, fraction, **kwargs):
        self.fraction = fraction
        super().__init__(**kwargs)

    @property
    def osip_description(self):
        return 'OSIPFrac' + str(self.fraction)

    @u.quantity_input(chan_mid_nominal=u.keV, equivalencies=u.spectral())
    def osip_range(self, chan_mid_nominal, order):
        return self.ccd_redist.interval(self.fraction, loc=chan_mid_nominal)


class FractionalDistanceOSIP(OSIPBase):
    '''Modify ARF files to account for order-sorting effects

    This class implements order sorting and integration of
    the probabilities (OSIP) for order sorting regions that fill a fixed
    fraction of the energy space.
    This includes order-sorting of a photons list
    (not yet implmented) and  methods to modify ARF files
    to account for order-sorting. Different diffraction orders fall on
    the same physical space on the CCD. The CCD energy resolution can
    then be used to assign photons to specific grating
    orders. However, since the CCD energy resolution is finite, this
    assingment is not perfect. Some photons may fall outside of the
    nominal energy used for order sorting and thus be lost, other may
    fall onto the energy assgined tothe different order. This class
    modifies ARF files to correct for the integrated probability that a photon
    falls inside the order-sorting region.

    Parameters
    ----------
    fraction : float
        Fraction (between 0 and 1) of the space between orders that will be
        covered by the extration region. For a value of 1, order-sorting
        regions just touch and each photon will be assigned to exactly one
        order.
    offset_orders : list of int
        Offset from main order that is relevant.
    ccd_redist : callable
        Function that return the width the Gaussian sigma of the CCD
        resolution, given an input energy.
    '''
    def __init__(self, fraction=1., **kwargs):
        self.fraction = fraction
        super().__init__(**kwargs)

    @property
    def osip_description(self):
        return 'OSIPDist' + str(self.fraction)

    @u.quantity_input(chan_mid_nominal=u.keV, equivalencies=u.spectral())
    def osip_tab(self, chan_mid_nominal, order):
        energy = chan_mid_nominal.to(u.keV, equivalencies=u.spectral())
        dE = energy / abs(order)
        return np.broadcast_to(self.fraction / 2 * dE,
                               (2, len(chan_mid_nominal)), subok=True)
